local MINOR_VERSION = tonumber(("$Revision: 74 $"):match("%d+"))
HitsMode = LibStub("AceAddon-3.0"):NewAddon("HitsMode", "AceConsole-3.0", "AceEvent-3.0")
HitsMode.version = GetAddOnMetadata("HitsMode", "Version");
HitsMode.MINOR_VERSION = MINOR_VERSION;
----

local WHT = "|cffffffff";
local BLK = "|cff000000";
local YEL = "|cffffff20";
local RED = "|cffff2020";
local GRN = "|cff20ff20";
local BLU = "|cff2020ff";
local HIL = "|cfffbd284";

BINDING_HEADER_HITSMODE = "HitsMode";
BINDING_NAME_OPTIONSTOGGLE = "Show HitsMode Options Window";

local debug2 = true;

local L = nil;
local isEnabled = 0;
local coerceChat = 0;
local currentWindow = 0;
local optionsVisible = 0;
local currentPanel = 1;
local lockUpdateOptions = 0;
local pOptions = nil;
local currentColor = 0;
local inCombat = 0;
local loadingHelp = 0;
local fullyLoaded = false;

local damageTaken = 0;
local damageGiven = 0;
local healsTaken = 0;
local healsGiven = 0;
local damageResisted = 0;
local petDamageGiven = 0;
local petDamageTaken = 0;
local petHealsTaken = 0;
local petHealsGiven = 0;
local manaGained = 0;
local energyGained = 0;
local rageGained = 0;
local attacksGained = 0;
local skillsIncreased = 0;
local expGained = 0;
local combatStartTime = 0;
local announced = false;

local g_debugWrote;

local NUMTABS = 6;
local PANELINDEX_CHARS = 5;
local PANELINDEX_PRESETS = 6;
local NUMOPTIONS = 26;
local OPTIONHEIGHT = 18;
local HM_EXPTYPE_EXPERIENCE = 1;
local HM_EXPTYPE_HONOR = 2;
local HM_EXPTYPE_REPUTATION = 3;

local SPELL_POWER_MANA = SPELL_POWER_MANA;
local SPELL_POWER_RAGE = SPELL_POWER_RAGE;
local SPELL_POWER_ENERGY = SPELL_POWER_ENERGY;
local SPELL_POWER_FOCUS = SPELL_POWER_FOCUS;
local SPELL_POWER_HAPPINESS = SPELL_POWER_HAPPINESS;
local SPELL_POWER_RUNES = SPELL_POWER_RUNES;
local SPELL_POWER_RUNIC_POWER = SPELL_POWER_RUNIC_POWER;
local SCHOOL_MASK_NONE = SCHOOL_MASK_NONE
local SCHOOL_MASK_PHYSICAL = SCHOOL_MASK_PHYSICAL
local SCHOOL_MASK_HOLY = SCHOOL_MASK_HOLY
local SCHOOL_MASK_FIRE = SCHOOL_MASK_FIRE
local SCHOOL_MASK_NATURE = SCHOOL_MASK_NATURE
local SCHOOL_MASK_FROST = SCHOOL_MASK_FROST
local SCHOOL_MASK_SHADOW = SCHOOL_MASK_SHADOW
local SCHOOL_MASK_ARCANE = SCHOOL_MASK_ARCANE

local HM_COMBATLOG_OBJECT_YOU = bit.bor(COMBATLOG_OBJECT_AFFILIATION_MINE, COMBATLOG_OBJECT_REACTION_FRIENDLY, COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_TYPE_PLAYER);
local HM_COMBATLOG_OBJECT_PARTYMEMBER = bit.bor(COMBATLOG_OBJECT_AFFILIATION_PARTY, COMBATLOG_OBJECT_REACTION_FRIENDLY, COMBATLOG_OBJECT_CONTROL_PLAYER, COMBATLOG_OBJECT_TYPE_PLAYER);


function HitsMode:OnEnable()
	-- Get localized strings
 L = self:Localize();
 if L == nil then return; end
 self:LocalizeXml();

 -- Register events
 local events = {
		"PLAYER_REGEN_DISABLED",
		"PLAYER_REGEN_ENABLED",
		"CHAT_MSG_COMBAT_XP_GAIN",
		"CHAT_MSG_LOOT",
		"CHAT_MSG_COMBAT_FACTION_CHANGE",
		"CHAT_MSG_COMBAT_HONOR_GAIN",
		"CHAT_MSG_MONEY",
		"CHAT_MSG_SKILL",
	};
 for key, value in pairs(events) do
  self:RegisterEvent(value, "OnEvent");
 end

 -- Add Slash Commands
 self:RegisterChatCommand("hitsmode", "ShowOptions");
 self:RegisterChatCommand("hm", "ShowOptions");
 
 -- Add extra font heights to the game
 CHAT_FONT_HEIGHTS = { 6, 8, 10, 11, 12, 14, 16, 18, 20, 24, 36, 48, 72, 96 };

 -- Startup the options system
 self:CreateOptions(L);
 self:InitializeOptionsSystem();

 -- Turn HitsMode on if it's enabled
 if (self:Get("enable") ~= 0) then
  self:TurnOn();
 end

 -- Initialize the options window
 PanelTemplates_SetNumTabs(HMOptionsPaneBgFrame, NUMTABS);
 self:SwitchPanel(1);

		-- Setup the Save Preset static popup
	StaticPopupDialogs["HITSMODE_SAVE_PRESET"] = {
		text = L[150], -- Enter the preset name
		button1 = TEXT(ACCEPT),
		button2 = TEXT(CANCEL),
		hasEditBox = 1,
		maxLetters = 60,
		whileDead = 1,
		OnAccept = function(renameID)
			local name = getglobal(this:GetParent():GetName().."EditBox"):GetText();
			getglobal(this:GetParent():GetName().."EditBox"):SetText("");
			self:SavePreset(name);
		end,
		timeout = 0,
		EditBoxOnEnterPressed = function(renameID)
			local name = getglobal(this:GetParent():GetName().."EditBox"):GetText();
			getglobal(this:GetParent():GetName().."EditBox"):SetText("");
			self:SavePreset(name);
			this:GetParent():Hide();
		end,
		EditBoxOnEscapePressed = function ()
			getglobal(this:GetParent():GetName().."EditBox"):SetText("");
			this:GetParent():Hide();
		end,
		hideOnEscape = 1
	};
	
	-- Cleanup
	self:Cleanup();

 -- Print our loaded message
 self:Print(string.format(L[159], self:C(.3, .3, 1), self:C(1, 1, 1), self.version, self:C(.3, .3, 1))); --**--
 fullyLoaded = true;
end


function HitsMode:Localize()
	local loc = GetLocale();
	if type(self["Localize_"..loc]) == "function" then
		return self["Localize_"..loc](self);
	elseif type(self["Localize_enUS"]) == "function" then
		return self:Localize_enUS();
	else
		self:Print("Error: HitsMode was unable to locate any localizations. Disabling.");
		return nil;
	end
end


function HitsMode:LocalizeXml()
	-- Localize all static UI strings
	local i;
	for i = 1, NUMOPTIONS do
  local prefix = "HMOptionLine"..i;
  getglobal(prefix.."_Load"):SetText(L[151]); -- Load
	end
	getglobal("HMSavePresetButton"):SetText(L[152]); -- Save Preset
	getglobal("HMOptionsPaneBgFrameTab1"):SetText(L[1]); -- General
	getglobal("HMOptionsPaneBgFrameTab2"):SetText(L[153]); -- Messages
	getglobal("HMOptionsPaneBgFrameTab3"):SetText(L[154]); -- Filters
	getglobal("HMOptionsPaneBgFrameTab4"):SetText(L[155]); -- Summary
	getglobal("HMOptionsPaneBgFrameTab5"):SetText(L[161]); -- Chars
	getglobal("HMOptionsPaneBgFrameTab6"):SetText(L[162]); -- Presets
	getglobal("HMColorPickerCopyButton"):SetText(L[163]); -- Copy
	getglobal("HMColorPickerPasteButton"):SetText(L[176]); -- Paste
	
	-- Force a resize on the tabs
	for i = 1, NUMTABS do
		local name = "HMOptionsPaneBgFrameTab"..i;
		local tab = getglobal(name);
		PanelTemplates_TabResize(tab, 0);
		getglobal(name.."HighlightTexture"):SetWidth(tab:GetTextWidth() + 31);
	end
end


function HitsMode:OnUpdateFuBarTooltip()
	GameTooltip:AddLine("|cffffff00"..L[287])
end

function HitsMode:OnFuBarClick(button)
	HitsMode:ToggleOptions();
end


function HitsMode:Cleanup()
	self.OptionsList = 0;
	self.OptionsList = nil;
	collectgarbage("collect");
end


function HitsMode:TurnOn()
 local c;
 
 if (isEnabled == 0) then
  self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", "OnEvent");
  coerceChat = 0;
  for c = 1, 4 do
   if (self:Get("enable"..c) == 1) then
    currentWindow = c;
    self:EnableMultiwindow(c);
   end
  end
  isEnabled = 1;
 end
end


function HitsMode:EnableMultiwindow(window)
 local shown, chatFrame, chatTab, _;
 local i = self:GetSliderValue("window"..window);
  
 _, _, _, _, _, _, shown, _ = GetChatWindowInfo(i);
 chatFrame = getglobal("ChatFrame"..i);
 chatTab = getglobal("ChatFrame"..i.."Tab");
 if (not shown and not chatFrame.isDocked and (i ~= 1)) then
  FCF_SetWindowName(chatFrame, "HitsMode "..window);
  FCF_SetWindowColor(chatFrame, DEFAULT_CHATFRAME_COLOR.r, DEFAULT_CHATFRAME_COLOR.g, DEFAULT_CHATFRAME_COLOR.b);
  FCF_SetWindowAlpha(chatFrame, DEFAULT_CHATFRAME_ALPHA);
  SetChatWindowLocked(i, nil);
  ChatFrame_RemoveAllMessageGroups(chatFrame);
  chatFrame:Show();
  SetChatWindowShown(i, 1);
  FCF_DockFrame(chatFrame, i);
 end
 self:PrintW(string.format(L[160], window)); -- HM window %s is here!
end


function HitsMode:TurnOff()
 if (isEnabled == 1) then
  self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
  isEnabled = 0;
 end
end


function HitsMode:ShowHelp()
 ToggleHelpFrame();
 loadingHelp = 1;
end


function HitsMode:HelpOnLoad(frame)
	LoadMicroButtonTextures(frame, "Help");
end


-- This is only called while the Options window is open and if the
-- help button is present. We need to use this to tell the
-- knowledge base to search for our help articles after it has loaded.
function HitsMode:HelpOnUpdate()
 if (loadingHelp == 1) then
  if (KBSetup_IsLoaded()) then
   loadingHelp = 0;
   KnowledgeBaseFrameEditBox:SetText("");
   UIDropDownMenu_SetSelectedName(KnowledgeBaseFrameCategoryDropDown, "HitsMode");
   KnowledgeBaseFrame_Search(1);
  end
 end
end


-- Main event handler
function HitsMode:OnEvent(event, ...)
 if (isEnabled) then
  self:HandleCombatEvent(event, ...);
 end
end


function HitsMode:Is(flags, flag)
	if (flags == nil or flag == nil) then return false; end
 return (bit.band(flags, flag) ~= 0);
end

function HitsMode:PackColor(r, g, b)
	return bit.lshift(bit.mod(r * 255, 256), 16) + bit.lshift(bit.mod(g * 255, 256), 8) + bit.mod(b * 255, 256);
end

function HitsMode:UnpackColor(c)
	return bit.band(bit.rshift(c, 16), 255) / 255, bit.band(bit.rshift(c, 8), 255) / 255, bit.band(c, 255) / 255;
end

function HitsMode:PackValues(a, b, c, d)
	local v = 0;
	if (a and a ~= 0) then v = v + 1; end
	if (b and b ~= 0) then v = v + 2; end
	if (c and c ~= 0) then v = v + 4; end
	if (d and d ~= 0) then v = v + 8; end
	return v;
end

function HitsMode:UnpackValues(v)
	return (bit.band(v, 1) > 0), (bit.band(v, 2) > 0), (bit.band(v, 4) > 0), (bit.band(v, 8) > 0);
end


-- Adds a message to the current HitsMode window
function HitsMode:PrintW(msg)
 local c = coerceChat;
 coerceChat = 0; -- turn off coerce so we make sure to get the right chat window during GetValue
 if (currentWindow == 0 or currentWindow == nil) then return; end
 getglobal("ChatFrame"..self:GetSliderValue("window"..currentWindow)):AddMessage(msg);
 coerceChat = c; -- set coerce back where it was
end


-- BuildColorCode:
-- Returns a string that changes the color in any WoW displayed string.
-- You can use |r to close the color change or just change to the next color.
function HitsMode:C(r, g, b)
 return string.format("|cff%02x%02x%02x", (r*255), (g*255), (b*255));
end


-- GetColorCode: Gets the configured color and converts it to a chat color code 
function HitsMode:GC(option)
 return self:C(self:GetColor(option));
end


-- Gets a configured chat color code and scales it by a certain value
function HitsMode:GetScaledColor(option, scale)
	local r, g, b = self:GetColor(option);
 return self:C(r * scale, g * scale, b * scale);
end


-- GetColorCodeOverlay:
-- Gets an overlayed chat color code. Requires some extra input, where:
--   unit is either you, pet, or mob
--   msgtype is crit or anything else
-- Crit overlay takes precedence; a Pet Crit will appear with Crit overlay only.
function HitsMode:GCO(option, unit, msgtype)
	local r, g, b = self:GetColor(option);
 
 if (msgtype == "crit") then
  if (self:Get("critoverlay") == 1) then
   local r2, g2, b2 = self:GetColor("critoverlay");
   r = (r / 2) + (r2 / 2);
   g = (g / 2) + (g2 / 2);
   b = (b / 2) + (b2 / 2);
  end
 elseif (unit == "pet") then
  if (self:Get("petoverlay") == 1) then
   local r2, g2, b2 = self:GetColor("critoverlay");
   r = (r / 2) + (r2 / 2);
   g = (g / 2) + (g2 / 2);
   b = (b / 2) + (b2 / 2);
  end
 end
 
 return self:C(r, g, b);
end


-- Gets the requested chat message string. This does the checking for capitalize.
function HitsMode:GetString(option)
 local s = self.ChatMessages[option];
 if (s == nil) then
  self:Print(string.format("[Error] GetString: Unable to locate \"%s\"", option));
  return " ";
 end
 if (self:Get("caps") == 1 and option ~= "FOR") then
  s = string.upper(string.sub(s, 1, 1))..string.sub(s, 2);
 end
 return s;
end


-- Returns brackets if they have been turned on
function HitsMode:BracketLeft()
 if (self:Get("brackets") == 1) then
  return self:GC("brackets")..self:GetString("BRACKETLEFT").."|r";
 else
  return "";
 end
end


function HitsMode:BracketRight()
 if (self:Get("brackets") == 1) then
  return self:GC("brackets")..self:GetString("BRACKETRIGHT").."|r";
 else
  return "";
 end
end


function HitsMode:Round(base, precision)
 local m = math.pow(10, precision);
 local a = math.floor(base * m) / m;
 return a;
end


function HitsMode:FormatNumber(number)
 return self:Round(number, 2);
end


function HitsMode:IsPet(s, f)
	if (s == nil or f == nil) then return false; end
	if (self:Is(f, COMBATLOG_OBJECT_TYPE_PET) or self:Is(f, COMBATLOG_OBJECT_TYPE_GUARDIAN)) and
	   (self:Is(f, COMBATLOG_OBJECT_CONTROL_PLAYER)) and
	   (self:Is(f, COMBATLOG_OBJECT_REACTION_FRIENDLY)) and
	   (self:Is(f, COMBATLOG_OBJECT_AFFILIATION_MINE)) then return true; end
	if (UnitName("pet") == nil) then return false;	end
	local b = (s == UnitName("pet") and self:Is(f, COMBATLOG_OBJECT_AFFILIATION_MINE) and self:Is(f, COMBATLOG_OBJECT_REACTION_FRIENDLY) and self:Is(f, COMBATLOG_OBJECT_CONTROL_PLAYER) and self:Is(f, COMBATLOG_OBJECT_TYPE_PET));
 return b;
end


function HitsMode:IsTarget(s, f)
	if (s == nil or f == nil) then return false; end
	local b = (s == UnitName("target") and self:Is(f, COMBATLOG_OBJECT_TARGET));
 return b;
end


function HitsMode:IsYou(s, f)
	if (s == nil or f == nil) then return false; end
	local b = (s == UnitName("player") and self:Is(f, COMBATLOG_OBJECT_AFFILIATION_MINE) and self:Is(f, COMBATLOG_OBJECT_REACTION_FRIENDLY) and self:Is(f, COMBATLOG_OBJECT_CONTROL_PLAYER) and self:Is(f, COMBATLOG_OBJECT_TYPE_PLAYER));
 return b;
end


function HitsMode:IsParty(s, f)
	if (s == nil or f == nil) then return false; end
	if (self:Is(f, COMBATLOG_OBJECT_AFFILIATION_PARTY) and self:Is(f, COMBATLOG_OBJECT_CONTROL_PLAYER) and self:Is(f, COMBATLOG_OBJECT_REACTION_FRIENDLY) and self:Is(f, COMBATLOG_OBJECT_TYPE_PLAYER)) then
		local i;
		for i = 1, 4 do
			if (s == UnitName("party"..i)) then return true; end
		end
	end
	return false;
end


function HitsMode:IsPartyPet(s, f)
	if (s == nil or f == nil) then return false; end
	if (self:Is(f, COMBATLOG_OBJECT_AFFILIATION_PARTY) and self:Is(f, COMBATLOG_OBJECT_CONTROL_PLAYER) and self:Is(f, COMBATLOG_OBJECT_REACTION_FRIENDLY) and self:Is(f, COMBATLOG_OBJECT_TYPE_PET)) then
		local i;
		for i = 1, 4 do
			if (s == UnitName("partypet"..i)) then return true; end
		end
	end
	return false;
end


function HitsMode:IsRaid(s, f)
	if (s == nil or f == nil) then return false; end
	if (self:Is(f, COMBATLOG_OBJECT_AFFILIATION_RAID) and self:Is(f, COMBATLOG_OBJECT_CONTROL_PLAYER) and self:Is(f, COMBATLOG_OBJECT_REACTION_FRIENDLY) and self:Is(f, COMBATLOG_OBJECT_TYPE_PLAYER)) then
		local i;
		for i = 1, 40 do
			if (s == UnitName("raid"..i)) then return true; end
		end
	end
	return false;
end


function HitsMode:IsRaidPet(s, f)
	if (s == nil or f == nil) then return false; end
	if (self:Is(f, COMBATLOG_OBJECT_AFFILIATION_RAID) and self:Is(f, COMBATLOG_OBJECT_CONTROL_PLAYER) and self:Is(f, COMBATLOG_OBJECT_REACTION_FRIENDLY) and self:Is(f, COMBATLOG_OBJECT_TYPE_PET)) then
		local i;
		for i = 1, 40 do
			if (s == UnitName("raidpet"..i)) then return true; end
		end
	end
	return false;
end


function HitsMode:IsInList(s, list)
	if (s == nil or list == nil) then return false; end
	local a, l, x;
	a = string.lower(string.gsub(s, "%s", ""));
	l = string.lower(list)..",";
 for x in string.gmatch(l, "(.-)%,") do
		x = string.gsub(x, "%s", "");
		--self:DebugWrite(string.format("checking name from comma list: %s against: %s", x, a));
		if (x == a) then return true; end
 end
 return false;
end


function HitsMode:IsCrit(critical)
	return (critical ~= nil);
end


function HitsMode:Damage(amount, critical)
	if amount == nil then return ""; else return string.format("%s", amount); end
end


function HitsMode:FormatBool(b)
	if b == true then return "true"; else return "false"; end
end


function HitsMode:Flip(a, b)
	return b, a;
end


function HitsMode:Add(a, b)
	if a == nil then a = 0; end
	if b == nil then b = 0; end
	return a + b;
end


function HitsMode:DebugWrite(s)
 local count;
 
 for count = 1, 4 do
  if (self:Get("enable"..count) == 1) then
   if (self:Get("debug", count) == 1) then
    getglobal("ChatFrame"..self:GetSliderValue("window"..count)):AddMessage(s);
	   g_debugWrote = true;
   end
  end
 end
end


function HitsMode:DebugWrite2(s)
	if (debug2 == true) then
	 self:DebugWrite(s);
	 g_debugWrote = true;
	end
end


function HitsMode:GetSpellIcon(spellId)
 local name, rank, icon, cost, isFunnel, powerType, castTime, minRange, maxRange = GetSpellInfo(spellId);
 return icon;
end


function HitsMode:GetItemLink(itemId)
	local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(itemId);
	return itemLink;
end


-- This system is very nil-safe, except for one assumption. Because of the order of
-- calls, the option system assumes that if a variable exists, it's entire structure
-- has been created. Therefore, it's important you properly initialize the option
-- system before any calls to Get or Set functions.
--
-- The structure of an option variable is as follows:
--
--   HM_Options["Player of Realm"]  (pOptions is set to this on init)
--     ["variable name"]
--       ["value"]  the single value of the option (also the value of sliders)
--       ["window"]
--  1..4    the individual toggles for each chat window
--       ["color"]
--  r, g, b  the color of the option (0-1 range)
--
-- This system relies on the self.OptionsList object to load default values. If a
-- default is not present, a 0 is initialized into the field.


-- Starts up the options system
function HitsMode:InitializeOptionsSystem()
 if (HM_Options == nil) then
  self:InitBlankOptions();
 end
 pOptions = self:GetOptionsPlayer();
end


function HitsMode:Get(var, num)
 if (coerceChat == 1) then num = currentWindow; end
 if (num == nil) then num = 1; end
 if (pOptions[var]) then
		a, b, c, d = self:UnpackValues(pOptions[var][1]);
		if (num == 1) then if (a) then return 1; else return 0; end
		elseif (num == 2) then if (b) then return 1; else return 0; end
		elseif (num == 3) then if (c) then return 1; else return 0; end
		elseif (num == 4) then if (d) then return 1; else return 0; end
		else
			self:Print(string.format("[Error] GetValue: Invalid window number %s specified for variable %s", num, var));
			return 0;
		end
	else
  self:Print(string.format("[Error] GetValue: Unable to locate \"%s\"", var));
  return 0;
	end
end

function HitsMode:SetValue(var, value, num)
 if (pOptions[var]) then
		a, b, c, d = self:UnpackValues(pOptions[var][1]);
		if (num == 1) then a = value;
		elseif (num == 2) then b = value;
		elseif (num == 3) then c = value;
		elseif (num == 4) then d = value;
		else
			self:Print(string.format("[Error] SetValue: Invalid window number %s specified for variable %s", num, var));
		end
		pOptions[var][1] = self:PackValues(a, b, c, d);
 else
  self:Print(string.format("[Error] SetValue: Unable to locate \"%s\"", var));
 end
end

function HitsMode:SetTextValue(var, value)
 if (pOptions[var]) then
		pOptions[var][4] = value;
	else
  self:Print(string.format("[Error] SetTextValue: Unable to locate \"%s\"", var));
 end
end

function HitsMode:GetTextValue(var)
 if (pOptions[var]) then
		return pOptions[var][4] or "";
	else
  self:Print(string.format("[Error] GetTextValue: Unable to locate \"%s\"", var));
  return "";
 end
end

function HitsMode:SetAllValues(var, value)
 self:SetValue(var, value, 1);
 self:SetValue(var, value, 2);
 self:SetValue(var, value, 3);
 self:SetValue(var, value, 4);
end

function HitsMode:SetSliderValue(var, value)
 if (pOptions[var]) then
		pOptions[var][3] = value;
	else
  self:Print(string.format("[Error] SetSliderValue: Unable to locate \"%s\"", var));
 end
end

function HitsMode:GetSliderValue(var, value)
 if (pOptions[var]) then
		return pOptions[var][3] or 0;
	else
  self:Print(string.format("[Error] GetSliderValue: Unable to locate \"%s\"", var));
 end
end

function HitsMode:GetColor(var)
 if (pOptions[var]) then
		return self:UnpackColor(pOptions[var][2] or 0);
 else
  self:Print(string.format("[Error] GetColor: Unable to locate \"%s\"", var));
  return 1, 1, 1;
 end
end

function HitsMode:SetColor(var, value)
 if (pOptions[var]) then
  pOptions[var][2] = self:PackColor(value.r, value.g, value.b);
 else
  self:Print(string.format("[Error] SetColor: Unable to locate \"%s\"", var));
 end
end


-- Creates the initial blank object
function HitsMode:InitBlankOptions()
 HM_Options = { };
end


-- Returns a pointer to the current player's option set, or creates a default if none present
function HitsMode:GetOptionsPlayer()
 name = UnitName("player").." of "..GetCVar("realmName");

 -- No current options object found, load all defaults
 if (HM_Options[name] == nil) then
  self:LoadDefaultOptions(name);
  
 -- Upgrading from a version prior to 1.2, which introduced multiwindow support and requires a reset
 elseif (HM_Options[name].version == "1.0" or HM_Options[name].version == "1.1") then
  self:Print(self:C(.3, .3, 1)..L[164]); -- Upgrading from a version prior to 1.2. Reset required.
  self:LoadDefaultOptions(name);
  
 -- Different version found, migrate options
 elseif (HM_Options[name].version ~= self.version) then
  self:Print(self:C(.3, .3, 1)..L[165]); -- You are using a different version of HitsMode: Migrating options.
  self:MigrateOptions(HM_Options[name]);
  
 -- Same version, but we're still going to migrate silently in case something got corrupted
 else
		self:MigrateOptions(HM_Options[name], true);
 end
 return HM_Options[name];
end


-- Makes a deep copy of a given table
function HitsMode:CopyTable(t, lookup_table)
 local copy = {};
 for i,v in pairs(t) do
  if type(v) ~= "table" then
   copy[i] = v;
  else
   lookup_table = lookup_table or {};
   lookup_table[t] = copy;
   if lookup_table[v] then
    copy[i] = lookup_table[v]; -- We already copied this table, reuse the copy
   else
    copy[i] = self:CopyTable(v,lookup_table); -- Not yet copied, copy it
   end
  end
 end
 return copy;
end


function HitsMode:PopulatePresets()
 self.OptionsList[PANELINDEX_PRESETS] = { };
 self.OptionsList[PANELINDEX_PRESETS][1] = {
  type = "preset",
  presetText = self:C(1, 1, 0).."Default Settings|r",
  presetName = "$$default",
  delete = 0
 };
 if (HM_Options) then
  if (HM_Options["presets"] == nil) then HM_Options["presets"] = { }; end
  local i = 2;
  for k, v in pairs(HM_Options["presets"]) do
   if (strsub(k, 1, 2) ~= "$$") then
				self:MigrateOptions(v, true);
    self.OptionsList[PANELINDEX_PRESETS][i] = {
     type = "preset",
     presetText = k,
     presetName = k
    };
    i = i + 1;
   end
  end
 end
end


function HitsMode:PopulateCharacters()
 local name = UnitName("player").." of "..GetCVar("realmName");
 local i = 1;
 self.OptionsList[PANELINDEX_CHARS] = { };
 if (HM_Options) then
  for k, v in pairs(HM_Options) do
   if (k ~= name and k ~= "global" and k ~= "presets" and v["version"] ~= "1.0" and v["version"] ~= "1.1") then
				self:MigrateOptions(v, true);
    self.OptionsList[PANELINDEX_CHARS][i] = {
     type = "preset",
     presetText = k,
     presetName = "$$"..k,
     delete = 0
    };
    i = i + 1;
   end
  end
  self.OptionsList["global"] = nil; -- Ditch legacy "global" settings
 end
end


function HitsMode:SavePresetOnClick()
 StaticPopup_Show("HITSMODE_SAVE_PRESET");
end


function HitsMode:SavePreset(presetName)
 if (strsub(presetName, 1, 2) ~= "$$") then
  local name = UnitName("player").." of "..GetCVar("realmName");
  if HM_Options["presets"] == nil then HM_Options["presets"] = { }; end
  HM_Options["presets"][presetName] = self:CopyTable(HM_Options[name]);
  self:PopulatePresets();
  self:UpdateOptions();
  self:Print(string.format(self:C(.3, .3, 1)..L[174], presetName)); -- Saved preset: %s
 else
  self:Print(string.format(self:C(.3, .3, 1)..L[178])); -- Preset name may not start with $$
 end
end


function HitsMode:DeletePreset(presetName)
 HM_Options["presets"][presetName] = nil;
 self:PopulatePresets();
 self:UpdateOptions();
 self:Print(string.format(self:C(.3, .3, 1)..L[175], presetName)); -- Deleted preset: %s
end


function HitsMode:LoadPreset(presetName)
 local name = UnitName("player").." of "..GetCVar("realmName");
 pOptions = nil;
 
 -- Defaults
 if (presetName == "$$default") then
  self:LoadDefaultOptions(name);
  
 -- Another character
 elseif (strsub(presetName, 1, 2) == "$$") then
  local copyName = strsub(presetName, 3);
  if (HM_Options[copyName] == nil) then
   self:Print(string.format(self:C(.3, .3, 1)..L[170], copyName)); -- Unable to locate settings for character: %s
  else
   HM_Options[name] = self:CopyTable(HM_Options[copyName]);
   self:Print(string.format(self:C(.3, .3, 1)..L[171], copyName)); -- Copied options from %s
  end
 
 -- Preset -- cannot find
 elseif (HM_Options["presets"][presetName] == nil) then
  self:Print(string.format(self:C(.3, .3, 1)..L[172], presetName)); -- Unable to locate preset: %s
  
 -- Preset -- found
 else
  HM_Options[name] = self:CopyTable(HM_Options["presets"][presetName]);
  self:Print(string.format(self:C(.3, .3, 1)..L[173], presetName)); -- Loaded preset: %s
 end
 
 pOptions = HM_Options[name];
 self:UpdateOptions();
end


-- Creates a new default configuration for the current player
function HitsMode:LoadDefaultOptions(name, silent)
 local c, p;

 HM_Options[name] = nil;
 HM_Options[name] = { };
 p = HM_Options[name];
 p.version = self.version;
 p.maxdps = 0.0;

 for c = 1, NUMTABS-2 do
  for key, value in pairs(self.OptionsList[c]) do
			if (value.var) then
			
				p[value.var] = {};
			
				-- Checkboxes
				if (value.single and value.default) then p[value.var][1] = self:PackValues(value.default, value.default, value.default, value.default);
				else p[value.var][1] = self:PackValues(value.d1 or false, value.d2 or false, value.d3 or false, value.d4 or false); end
				
				-- Color
				if (value.hideColor ~= true and value.c) then p[value.var][2] = self:PackColor(value.c.r or 0, value.c.g or 0, value.c.b or 0); end
				
				-- Slider value
				if (value.type == "slider") then
					p[value.var][2] = 0; -- Fill in the color to keep array entries linear
					p[value.var][3] = value.default or 0;
				
				-- Text value
				elseif (value.type == "text") then
					p[value.var][2] = 0; -- Keep array entries linear again
					p[value.var][3] = 0; -- Still keeping array entries linear, can't skip any
					p[value.var][4] = value.defaultText or "";
				end
				
   end
  end
 end
 
 if (silent == nil) then
  self:Print(self:C(.3, .3, 1)..string.format(L[166], name)); -- Default options loaded for %s.
 end
end


-- Migrates options from an older version. Converts old options format to new format,
-- loads defaults for any missing options, and deletes and options that are no longer in use.
function HitsMode:MigrateOptions(p, silent)
 local c, count, oldversion, deleted, converted;

 oldversion = p.version or "(unknown)";
 count = 0;
 deleted = 0;
 converted = 0;

 -- Delete options that no longer exist
 for key, value in pairs(p) do
		local found = false;
		for c = 1, NUMTABS-2 do
			for key2, value2 in pairs(self.OptionsList[c]) do
				if (value2.var == key) then
					found = true;
					break;
				end
			end
		end
		if (found == false and key ~= "version" and key ~= "maxdps") then
			p[key] = nil;
			deleted = deleted + 1;
		end
	end
	
	-- Set the version
	p.version = self.version;

	-- Convert old format options to new format
	for c = 1, NUMTABS-2 do
  for key, value in pairs(self.OptionsList[c]) do
   if (value.var and p[value.var]) then
				if (p[value.var].value or p[value.var].window or p[value.var].color or p[value.var].textValue) then

					-- Checkboxes
					if (value.single and p[value.var].value) then p[value.var][1] = p[value.var].value;
					elseif (p[value.var].window) then p[value.var][1] = self:PackValues(p[value.var].window[1] or 0, p[value.var].window[2] or 0, p[value.var].window[3] or 0, p[value.var].window[4] or 0);
					else p[value.var][1] = 0; end

					-- Color
					if (p[value.var].color) then p[value.var][2] = self:PackColor(p[value.var].color.r or 0, p[value.var].color.g or 0, p[value.var].color.b or 0);
					else p[value.var][2] = self:PackColor(1, 1, 1); end

					-- Slider value
					if (value.type == "slider") then p[value.var][3] = p[value.var].value or 0; end

					-- Text value
					if (value.type == "text") then
						p[value.var][3] = 0;
						p[value.var][4] = p[value.var].textValue or "";
					end

					-- Drop old data
					p[value.var].value = nil;
					p[value.var].window = nil;
					p[value.var].color = nil;
					p[value.var].textValue = nil;

					converted = converted + 1;
				end
			end
		end	
	end

	-- Get Lua to reallocate the table to free up space
	p["resetTrigger"] = 0;
	p["resetTrigger"] = nil;

	-- Load any missing defaults
 for c = 1, NUMTABS-2 do
  for key, value in pairs(self.OptionsList[c]) do
   if (value.var) then
				if (p[value.var] == nil) then
				
					p[value.var] = {};

					-- Checkboxes
					if (value.single and value.default) then p[value.var][1] = self:PackValues(value.default, value.default, value.default, value.default);
					else p[value.var][1] = self:PackValues(value.d1 or false, value.d2 or false, value.d3 or false, value.d4 or false); end

					-- Color
					if (value.hideColor ~= true and value.c) then p[value.var][2] = self:PackColor(value.c.r or 1, value.c.g or 1, value.c.b or 1);
					else p[value.var][2] = self:PackColor(1, 1, 1); end

					-- Slider value
					if (value.type == "slider") then
						p[value.var][2] = 0; -- Fill in the color to keep array entries linear
						p[value.var][3] = value.default or 0;

					-- Text value
					elseif (value.type == "text") then
						p[value.var][2] = 0; -- Keep array entries linear again
						p[value.var][3] = 0; -- Still keeping array entries linear, can't skip any
						p[value.var][4] = value.defaultText or "";
					end

     count = count + 1;
    end
   end
  end
 end

 if (silent == nil) then
  self:Print(self:C(.3, .3, 1)..string.format(L[167], oldversion)); -- Options migrated from version %s.
  self:Print(self:C(.3, .3, 1)..string.format(L[168], count)); -- %s new default values loaded.
  self:Print(self:C(.3, .3, 1)..string.format(L[288], deleted)); -- %s old options deleted.
  self:Print(self:C(.3, .3, 1)..string.format(L[289], converted)); -- %s options converted to new format.
 end
end


-- Displays the options window
function HitsMode:ShowOptions()
	self:CreateOptions(L);
	HMHelpMicroButton.tooltipText = L[218]; -- Get help about how to use HitsMode.
 ShowUIPanel(HMOptions);
end


-- Hides the options window
function HitsMode:HideOptions()
 HideUIPanel(HMOptions);
 self:Cleanup();
end


-- Toggles the options window
function HitsMode:ToggleOptions()
	if (fullyLoaded ~= true or L == nil) then return; end
 if (optionsVisible == 1) then
  self:HideOptions();
 else
  self:ShowOptions();
 end
end


-- The OnShow handler
function HitsMode:OptionsFrameOnShow()
 self:TurnOff();
 PlaySound("igMainMenuOpen");
 optionsVisible = 1;
 self:UpdateOptions();
 --getglobal("HMOptionsWindowHeaderText"):SetText("HitsMode v"..self.version.." rev "..self.MINOR_VERSION);
 getglobal("HMOptionsWindowHeaderText"):SetText("HitsMode4");
end


-- The OnHide handler
function HitsMode:OptionsFrameOnHide()
 self:SetInfo("");
 PlaySound("igMainMenuClose");
 optionsVisible = 0;
 if (self:Get("enable") == 1) then
  self:TurnOn();
 end
end


function HitsMode:OptionsOnVerticalScroll(frame, offset)
	FauxScrollFrame_OnVerticalScroll(frame, offset, OPTIONHEIGHT, function() end);
	HitsMode:UpdateOptions();
end


-- Gets the number of options in the current panel
function HitsMode:GetOptionCount()
 return table.getn(self.OptionsList[currentPanel]);
end


-- Gets the current offset
function HitsMode:GetOffset()
 return FauxScrollFrame_GetOffset(HMOptionsPane);
end


-- Draws the option window
-- Important: the options system must be fully initialized before calling this
function HitsMode:UpdateOptions()
 local c, cc;
 local panel = currentPanel;
 local offset = self:GetOffset();
 local numItems = self:GetOptionCount();
 
 FauxScrollFrame_Update(HMOptionsPane, numItems, NUMOPTIONS, OPTIONHEIGHT);
 
 if (lockUpdateOptions == 0) then
  lockUpdateOptions = 1;
 else
  return;
 end
 
 for c = 1, NUMOPTIONS do
  local prefix = "HMOptionLine"..c;
  local line = getglobal(prefix);
  local highlight = getglobal(prefix.."_HighlightBox");
  local headerHighlight = getglobal(prefix.."_HeaderHighlightBox");
  local description = getglobal(prefix.."_Description");
  local color = getglobal(prefix.."_Color");
  local swatch = getglobal(prefix.."_ColorSwatchTexture");
  local heading = getglobal(prefix.."_Heading");
  local slider = getglobal(prefix.."_Slider");
  local divider = getglobal(prefix.."_DividerLine");
  local delete = getglobal(prefix.."_Delete");
  local load = getglobal(prefix.."_Load");
  local editbox = getglobal(prefix.."_EditBox");
  
  local config = self.OptionsList[panel][c+offset];
    
  -- An option exists
  if (config) then
   line:Show();
   self:HideOptionsLine(c);
   if (c == NUMOPTIONS) then
    divider:Hide();
   end

   -- Heading
   if (config.type == "heading") then
    line.disableHighlight = 1;
    heading:SetText(L[config.t]);
    heading:Show();
    headerHighlight:Show();
    if config.hideSelectAll == nil or config.hideSelectAll == false then
					for cc = 1, 4 do
						local checkbutton = getglobal("HMOptionLine"..c.."_CheckButton"..cc);
						local i = c + offset + 1;
						local checked = true;
						while self.OptionsList[panel][i] ~= nil and self.OptionsList[panel][i].type ~= "heading" do
							if self:Get(self.OptionsList[panel][i].var, cc) == 0 then
								checked = false;
								break;
							end
							i = i + 1;
						end
						checkbutton:SetChecked(checked);
						checkbutton:Show();
					end
				end
    
   -- Slider
   elseif (config.type == "slider") then
    local string = getglobal(prefix.."_SliderText");
    local low = getglobal(prefix.."_SliderLow");
    local high = getglobal(prefix.."_SliderHigh");
    local valuetext = getglobal(prefix.."_SliderValueText");
    local getvalue = self:GetSliderValue(config.var);
    slider:SetMinMaxValues(config.minValue, config.maxValue);
    slider:SetValueStep(config.valueStep);
    slider:SetValue(getvalue);
    valuetext:SetText(getvalue);
    description:SetText(L[config.t]);
    low:SetText(config.minText);
    high:SetText(config.maxText);
    description:Show();
    slider:Show();
    
   -- Spell school
   elseif (config.type == "school") then
    description:SetText(L[config.t]);
    description:Show();
    if (config.hideColor == nil) then
     local r, g, b = self:GetColor(config.var);
     color.r = r;
     color.g = g;
     color.b = b;
     swatch:SetTexture(r, g, b);
     color:Show();
    end
    
   -- Preset 
   elseif (config.type == "preset") then
    description:SetText(config.presetText);
    description:Show();
    if (config.delete == nil or config.delete == 1) then delete:Show(); end
    if (config.show == nil or config.load == 1) then load:Show(); end
    line.presetName = config.presetName;
    
   -- Text entry option
   elseif (config.type == "text") then
    for cc = 1, 4 do
     local checkbutton = getglobal("HMOptionLine"..c.."_CheckButton"..cc);
     checkbutton:SetChecked(self:Get(config.var, cc));
     if (config.single == nil) then
      checkbutton:Show();
     elseif (cc == 4) then
      checkbutton:Show();
      checkbutton:SetChecked(self:Get(config.var));
     end
    end
    description:SetText(L[config.t]);
    description:Show();
    line.disableHighlight = nil;
    editbox:Show();
    editbox:SetText(self:GetTextValue(config.var));

   -- Regular option
   else
    for cc = 1, 4 do
     local checkbutton = getglobal("HMOptionLine"..c.."_CheckButton"..cc);
     checkbutton:SetChecked(self:Get(config.var, cc));
     if (config.single == nil) then
      checkbutton:Show();
     elseif (cc == 4) then
      checkbutton:Show();
      checkbutton:SetChecked(self:Get(config.var));
     end
    end
    description:SetText(L[config.t]);
    description:Show();
    line.disableHighlight = nil;
    if (config.hideColor == nil) then
     local r, g, b = self:GetColor(config.var);
     color.r = r;
     color.g = g;
     color.b = b;
     swatch:SetTexture(r, g, b);
     color:Show();
    end    
    
   end

  -- No option exists
  else
   line:Hide();
  end
 end
 
 lockUpdateOptions = 0;
end


function HitsMode:HideOptionsLine(c)
 local cc;
 local prefix = "HMOptionLine"..c;
 local line = getglobal(prefix);
 local highlight = getglobal(prefix.."_HighlightBox");
 local headerHighlight = getglobal(prefix.."_HeaderHighlightBox");
 local description = getglobal(prefix.."_Description");
 local color = getglobal(prefix.."_Color");
 local heading = getglobal(prefix.."_Heading");
 local slider = getglobal(prefix.."_Slider");
 local delete = getglobal(prefix.."_Delete");
 local load = getglobal(prefix.."_Load");
 local editbox = getglobal(prefix.."_EditBox");

 for cc = 1, 4 do
  local checkbutton = getglobal("HMOptionLine"..c.."_CheckButton"..cc);
  checkbutton:Hide();
 end
 description:Hide();
 color:Hide();
 highlight:Hide();
 heading:Hide();
 slider:Hide();
 headerHighlight:Hide();
 load:Hide();
 delete:Hide();
 line.disableHighlight = nil;
 line.presetName = nil;
 line.presetDisplayName = nil;
 editbox:Hide();
end


function HitsMode:SwitchPanel(panel)
 currentPanel = panel;
 FauxScrollFrame_SetOffset(HMOptionsPane, 0);
 getglobal("HMOptionsPaneBgFrameTab"..panel):SetFrameLevel(HMOptionsPaneBgFrame:GetFrameLevel() + 1);
 PanelTemplates_SetTab(HMOptionsPaneBgFrame, panel);
 getglobal("HMOptionsPaneScrollBar"):SetValue(0);
 self:UpdateOptions();
end


function HitsMode:SetInfo(str)
 if (string == "" or string == nil) then
  GameTooltip:Hide();
 else
  GameTooltip:SetOwner(UIParent, "ANCHOR_BOTTOMRIGHT");
  GameTooltip:SetText(str);
  GameTooltip:Show();
 end
end


function HitsMode:OnMouseEnter(o)
 if (o:GetName() == "HMSavePresetButton") then
  self:SetInfo(L[177]);
 else
  if (getglobal(o:GetName()).disableHighlight == nil) then
   if (getglobal(o:GetName().."_HighlightBox") ~= nil) then
    getglobal(o:GetName().."_HighlightBox"):Show();
   end
  end
  local offset = tonumber(o:GetID());
  local option = self.OptionsList[currentPanel][self:GetOffset()+offset];
  if (option) then
   if (option.tip) then
    self:SetInfo(L[option.tip]);
   else
    self:SetInfo("");
   end
  else
   self:SetInfo("");
  end
 end
end


function HitsMode:OnMouseLeave(o)
 if (getglobal(o:GetName().."_HighlightBox")) then
  if (getglobal(o:GetName()).disableHighlight == nil) then
   getglobal(o:GetName().."_HighlightBox"):Hide();
  end
 end
 self:SetInfo("");
end


function HitsMode:SliderOnValueChanged(o)
 local slider = getglobal(o:GetName().."_Slider");
 local offset = tonumber(o:GetID());
 local option = self.OptionsList[currentPanel][self:GetOffset()+offset];
 self:SetSliderValue(option.var, slider:GetValue());
end


function HitsMode:CheckboxOnClick(o)
 local offset = tonumber(o:GetParent():GetID());
 local option = self.OptionsList[currentPanel][self:GetOffset()+offset];
 local checked;
 if (o:GetChecked()) then
  checked = 1;
 else
  checked = 0;
 end
 if option.type == "heading" then
		local i = self:GetOffset() + offset + 1;
		while self.OptionsList[currentPanel][i] ~= nil and self.OptionsList[currentPanel][i].type ~= "heading" do
			if (self.OptionsList[currentPanel][i].single) then
				self:SetAllValues(self.OptionsList[currentPanel][i].var, checked);
			else
				self:SetValue(self.OptionsList[currentPanel][i].var, checked, tonumber(o:GetID()));
			end
			i = i + 1;
		end
 else
		self:SetValue(option.var, checked, tonumber(o:GetID()));
		if (option.single) then
			self:SetAllValues(option.var, checked);
		end
		--if (option.var == "verbosea" and checked == 1) then
			--self:SetValue("verboseb", 0, tonumber(o:GetID()));
			--self:SetValue("verbosec", 0, tonumber(o:GetID()));
		--if (option.var == "verboseb" and checked == 1) then
			--self:SetValue("verbosea", 0, tonumber(o:GetID()));
			--self:SetValue("verbosec", 0, tonumber(o:GetID()));
		--elseif (option.var == "verbosec" and checked == 1) then
			--self:SetValue("verbosea", 0, tonumber(o:GetID()));
			--self:SetValue("verboseb", 0, tonumber(o:GetID()));
		--end
 end
	self:UpdateOptions();
end


function HitsMode:EditBoxOnTextChanged(o)
 local offset = tonumber(o:GetParent():GetID());
 local option = self.OptionsList[currentPanel][self:GetOffset()+offset];
 self:SetTextValue(option.var, o:GetText());
end


function HitsMode:OpenColorPicker(c, o)
	currentColor = c;
 CloseMenus();
 local offset = tonumber(o:GetID());
 local option = self.OptionsList[currentPanel][self:GetOffset()+offset];
 local prefix = "HMOptionLine"..currentColor;
 local swatch = getglobal(prefix.."_ColorSwatchTexture");
 local color = getglobal(prefix.."_Color");
 local r, g, b = self:GetColor(option.var);
 color.r = r;
 color.g = g;
 color.b = b;
 HMColorPicker:SetColorRGB(color.r, color.g, color.b);
 HMColorPicker.previousValues = { r = color.r, g = color.g, b = color.b };
 HMColorPicker:Show();
end


function HitsMode:ColorFunc()
 local r, g, b = HMColorPicker:GetColorRGB();
 local offset = tonumber(currentColor);
 local option = self.OptionsList[currentPanel][self:GetOffset()+offset];
 local prefix = "HMOptionLine"..currentColor;
 local swatch = getglobal(prefix.."_ColorSwatchTexture");
 local color = getglobal(prefix.."_Color");
 local savedColor = { };
 
 swatch:SetTexture(r, g, b);
 color.r = r;
 color.g = g;
 color.b = b;
 savedColor.r = r;
 savedColor.g = g;
 savedColor.b = b;
 self:SetColor(option.var, savedColor);

 if HM_ColorPicker_InTextChanged ~= true then
		HMColorPicker_ValueR:SetText((r*255));
		HMColorPicker_ValueG:SetText((g*255));
		HMColorPicker_ValueB:SetText((b*255));
 end
end


function HitsMode:ColorPickerOnTextChanged()
	HM_ColorPicker_InTextChanged = true;

	local rText = HMColorPicker_ValueR:GetText();
	if rText == "" or rText == nil then rText = "0" end;
	local r = tonumber(rText) / 255;
	if r > 1 then
		r = 1;
		HMColorPicker_ValueR:SetText((r*255));
	end
	local gText = HMColorPicker_ValueG:GetText();
	if gText == "" or gText == nil then gText = "0" end;
	local g = tonumber(gText) / 255;
	if g > 1 then
		g = 1;
		HMColorPicker_ValueG:SetText((g*255));
	end
	local bText = HMColorPicker_ValueB:GetText();
	if bText == "" or bText == nil then bText = "0" end;
	local b = tonumber(bText) / 255;
	if b > 1 then
		b = 1;
		HMColorPicker_ValueB:SetText((b*255));
	end

 local offset = tonumber(currentColor);
 local option = self.OptionsList[currentPanel][self:GetOffset()+offset];
 local prefix = "HMOptionLine"..currentColor;
 local swatch = getglobal(prefix.."_ColorSwatchTexture");
 local color = getglobal(prefix.."_Color");
 local savedColor = { };
 
	HMColorPicker:SetColorRGB(r, g, b);
	HMColorSwatch:SetTexture(r, g, b);
 swatch:SetTexture(r, g, b);
 color.r = r;
 color.g = g;
 color.b = b;
 savedColor.r = r;
 savedColor.g = g;
 savedColor.b = b;
 self:SetColor(option.var, savedColor);
 
 HM_ColorPicker_InTextChanged = false;
end


function HitsMode:ColorPickerCopy()
 HM_SavedR, HM_SavedG, HM_SavedB = HMColorPicker:GetColorRGB();
end


function HitsMode:ColorPickerPaste()
	if (HM_SavedR ~= nil and HM_SavedG ~= nil and HM_SavedB ~= nil) then
		HMColorPicker_ValueR:SetText((HM_SavedR*255));
		HMColorPicker_ValueG:SetText((HM_SavedG*255));
		HMColorPicker_ValueB:SetText((HM_SavedB*255));
		self:ColorPickerOnTextChanged();
	end
end


function HitsMode:CancelFunc()
 local offset = tonumber(currentColor);
 local option = self.OptionsList[currentPanel][self:GetOffset()+offset];
 local prefix = "HMOptionLine"..currentColor;
 local swatch = getglobal(prefix.."_ColorSwatchTexture");
 
 swatch:SetTexture(HMColorPicker.previousValues.r, HMColorPicker.previousValues.g, HMColorPicker.previousValues.b);
 self:SetColor(option.var, HMColorPicker.previousValues);
end


function HitsMode:FormatCombatEvent(...)
	 local timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags = select(1, ...);
	 local message = format("%s, %s, %s, %s, %s", eventtype or "nil", srcName or "nil", srcFlags or "nil", dstName or "nil", dstFlags or "nil");
  local pipe = false;
	 for i = 9, select("#", ...) do
	  if pipe == false then
	   pipe = true;
	   message = message.." || ";
	  else
	   message = message..", ";
	  end
		 message = message..(select(i, ...) or "nil");
	 end
	 return message;
end


--
-- The primary event handler for the combat log. This function retrieves the appropriate values from the
-- WoW event and passes them along to various output functions.
--
function HitsMode:HandleCombatEvent(event, ...)
 
 local arg1, arg2 = select(1, ...);
 local _, d, d2, name, spell, damage;
 
 -- COMBAT EVENTS -------------------------------------------------------------------------------------------------------------
 
 if (event == "COMBAT_LOG_EVENT_UNFILTERED") then
  self:DebugWrite(self:FormatCombatEvent(...));
 
  local timestamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags = select(1, ...);
	 local amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing, overhealing, missType, environmentalType;
	 local spellId, spellName, spellSchool, powerType, extraAmount, extraSpellId, extraSpellName, extraSpellSchool, auraType, extraSpell_verboseMessage;
	 local schoolName, powerName, extraSchoolName, itemId, itemName, amountMissed;

	 if (eventType == nil) then return; end
 	
	 -- DAMAGING EVENTS AND MISSES

	 if (eventType == "SWING_DAMAGE") then
		 amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing = select(9, ...);
		 amount = amount - (overkill or 0);
		 self:OutputDamage(eventType, srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, amount, school, resisted, blocked, absorbed, critical, glancing, crushing, environmentalType, overkill);

	 elseif (eventType == "SWING_MISSED") then
		 missType, amountMissed = select(9, ...);
   self:OutputMiss(eventType, srcName, srcFlags, dstName, dstFlags, missType, spellId, spellName, spellSchool, amountMissed);
   
	 elseif (eventType == "RANGE_DAMAGE") or (eventType == "SPELL_DAMAGE") or (eventType == "SPELL_PERIODIC_DAMAGE") or (eventType == "DAMAGE_SHIELD") or (eventType == "DAMAGE_SPLIT") or (eventType == "SPELL_BUILDING_DAMAGE") then
		 spellId, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing = select(9, ...);
		 amount = amount - (overkill or 0);
		 self:OutputDamage(eventType, srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, amount, school, resisted, blocked, absorbed, critical, glancing, crushing, environmentalType, overkill);

	 elseif (eventType == "RANGE_MISSED") or (eventType == "SPELL_MISSED") or (eventType == "SPELL_PERIODIC_MISSED") or (eventType == "DAMAGE_SHIELD_MISSED") then
		 spellId, spellName, spellSchool, missType, amountMissed = select(9, ...);
   self:OutputMiss(eventType, srcName, srcFlags, dstName, dstFlags, missType, spellId, spellName, spellSchool, amountMissed);

	 elseif (eventType == "ENVIRONMENTAL_DAMAGE") then
		 environmentalType, amount, overkill, school, resisted, blocked, absorbed, critical, glancing, crushing = select(9, ...);
		 amount = amount - (overkill or 0);
			self:OutputDamage(eventType, srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, amount, school, resisted, blocked, absorbed, critical, glancing, crushing, environmentalType, overkill);

  -- HEALING EVENTS
  
	 elseif (eventType == "SPELL_HEAL") or (eventType == "SPELL_PERIODIC_HEAL") or (eventType == "SPELL_BUILDING_HEAL") then
		 spellId, spellName, spellSchool, amount, overhealing, absorbed, critical = select(9, ...);
		 amount = amount - (overhealing or 0);
		 self:OutputHeal(eventType, srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, amount, critical, overhealing, absorbed);

  -- DEATH EVENTS
  
	 elseif (eventType == "PARTY_KILL") then
			if (self:IsYou(srcName, srcFlags)) then self:Output("mob_died", srcName, srcFlags, dstName, dstFlags, nil, "KILL");
			else self:Output("oth_kill", srcName, srcFlags, dstName, dstFlags, nil, "KILLS");
			end

  -- SUMMONING, CREATING, RESURRECTING
  
	 elseif (eventType == "SPELL_SUMMON") or (eventType == "SPELL_CREATE") or (eventType == "SPELL_RESURRECT") then
		 spellId, spellName, spellSchool = select(9, ...);
		 local verbose;
		 if (eventType == "SPELL_SUMMON") then
				if (self:IsYou(srcName, srcFlags)) then verbose = "SUMMONS"; else verbose = "SUMMONS"; end
			elseif (eventType == "SPELL_RESURRECT") then
				if (self:IsYou(srcName, srcFlags)) then verbose = "RESURRECT"; else verbose = "RESURRECTS"; end
			else
				if (self:IsYou(srcName, srcFlags)) then
				 if ((spellName or "") == "") then verbose = "CREATE"; else verbose = "CREATES"; end
				else
				 verbose = "CREATES";
				end
			end
		 if (self:IsYou(srcName, srcFlags)) then self:Output("you_summon", srcName, srcFlags, dstName, dstFlags, spellName, verbose, { spellSchool = spellSchool, spellId = spellId });
		 elseif (self:IsTarget(srcName, srcFlags)) then self:Output("mob_summon", srcName, srcFlags, dstName, dstFlags, spellName, verbose, { spellSchool = spellSchool, spellId = spellId });
		 else self:Output("oth_summon", srcName, srcFlags, dstName, dstFlags, spellName, verbose, { spellSchool = spellSchool, spellId = spellId });
		 end

  -- GAINS, DRAINS, LEECHES, EXTRA ATTACKS
  
	 elseif (eventType == "SPELL_ENERGIZE") or (eventType == "SPELL_PERIODIC_ENERGIZE") then
		 spellId, spellName, spellSchool, amount, powerType = select(9, ...);
			self:OutputEnergize(eventType, srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, amount, powerType);

	 elseif (eventType == "SPELL_DRAIN") or (eventType == "SPELL_PERIODIC_DRAIN") then
		 spellId, spellName, spellSchool, amount, powerType, extraAmount = select(9, ...);
		 self:OutputDrain(eventType, srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, amount, powerType, extraAmount);

	 elseif (eventType == "SPELL_LEECH") or (eventType == "SPELL_PERIODIC_LEECH") then
		 -- could be Dark Pact moving mana from an imp to its warlock
		 -- it seems that amount is the gain for the source and extraAmount is the loss for the dest
		 spellId, spellName, spellSchool, amount, powerType, extraAmount = select(9, ...);
			self:OutputEnergize(eventType, srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, amount, powerType);

	 elseif (eventType == "SPELL_EXTRA_ATTACKS") then
		 spellId, spellName, spellSchool, amount = select(9, ...);
		 self:OutputExtraAttacks(eventType, srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, amount);

  -- AURA EVENTS
  
	 elseif (eventType == "SPELL_AURA_APPLIED") or (eventType == "SPELL_AURA_REFRESH") or (eventType == "SPELL_AURA_REMOVED") then
		 spellId, spellName, spellSchool, auraType = select(9, ...);
			self:OutputAura(eventType, srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, auraType, nil);

	 elseif (eventType == "SPELL_AURA_APPLIED_DOSE") or (eventType == "SPELL_AURA_REMOVED_DOSE") then
		 spellId, spellName, spellSchool, auraType, amount = select(9, ...);
			self:OutputAura(eventType, srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, auraType, amount);
		
		-- DISPELS, SPELL STEALS, AURA BREAKS, INTERRUPTS

	 elseif (eventType == "SPELL_DISPEL") then
		 -- the first spell is the one causing the aura to be dispelled or otherwise cancelled, and the second is the name of the aura that is being dispelled
		 extraSpellId, extraSpell, extraSpellSchool, spellId, spellName, spellSchool, auraType = select(9, ...);
		 extraSpell_verboseMessage = "USING";
		 if (self:IsYou(srcName, srcFlags)) then self:Output("you_dispel", srcName, srcFlags, dstName, dstFlags, spellName, "DISPEL", { spellSchool = spellSchool, spellId = spellId, extraSpellId = extraSpellId, extraSpell = extraSpell, extraSpellSchool = extraSpellSchool, extraSpell_verboseMessage = extraSpell_verboseMessage });
		 elseif (self:IsPet(srcName, srcFlags)) then self:Output("pet_dispel", srcName, srcFlags, dstName, dstFlags, spellName, "DISPELS", { spellSchool = spellSchool, spellId = spellId, extraSpellId = extraSpellId, extraSpell = extraSpell, extraSpellSchool = extraSpellSchool, extraSpell_verboseMessage = extraSpell_verboseMessage });
		 elseif (self:IsTarget(srcName, srcFlags)) then self:Output("mob_dispel", srcName, srcFlags, dstName, dstFlags, spellName, "DISPELS", { spellSchool = spellSchool, spellId = spellId, extraSpellId = extraSpellId, extraSpell = extraSpell, extraSpellSchool = extraSpellSchool, extraSpell_verboseMessage = extraSpell_verboseMessage });
		 else self:Output("oth_dispel", srcName, srcFlags, dstName, dstFlags, spellName, "DISPELS", { spellSchool = spellSchool, spellId = spellId, extraSpellId = extraSpellId, extraSpell = extraSpell, extraSpellSchool = extraSpellSchool, extraSpell_verboseMessage = extraSpell_verboseMessage });
		 end

	 elseif (eventType == "SPELL_STOLEN") then
	  -- First spell is the one that caused the spell steal. Second spell is the spell that was stolen.
		 extraSpellId, extraSpell, extraSpellSchool, spellId, spellName, spellSchool, auraType = select(9, ...);
		 extraSpell_verboseMessage = "USING";
		 if (self:IsYou(srcName, srcFlags)) then self:Output("you_dispel", srcName, srcFlags, dstName, dstFlags, spellName, "STEAL", { spellSchool = spellSchool, spellId = spellId, extraSpellId = extraSpellId, extraSpell = extraSpell, extraSpellSchool = extraSpellSchool, extraSpell_verboseMessage = extraSpell_verboseMessage });
		 elseif (self:IsPet(srcName, srcFlags)) then self:Output("pet_dispel", srcName, srcFlags, dstName, dstFlags, spellName, "STEALS", { spellSchool = spellSchool, spellId = spellId, extraSpellId = extraSpellId, extraSpell = extraSpell, extraSpellSchool = extraSpellSchool, extraSpell_verboseMessage = extraSpell_verboseMessage });
		 elseif (self:IsTarget(srcName, srcFlags)) then self:Output("mob_dispel", srcName, srcFlags, dstName, dstFlags, spellName, "STEALS", nil, nil, extraSpellSchool, { spellSchool = spellSchool, spellId = spellId, extraSpellId = extraSpellId, extraSpell = extraSpell, extraSpellSchool = extraSpellSchool, extraSpell_verboseMessage = extraSpell_verboseMessage });
		 else self:Output("oth_dispel", srcName, srcFlags, dstName, dstFlags, spellName, "STEALS", { spellSchool = spellSchool, spellId = spellId, extraSpellId = extraSpellId, extraSpell = extraSpell, extraSpellSchool = extraSpellSchool, extraSpell_verboseMessage = extraSpell_verboseMessage });
		 end

	 elseif (eventType == "SPELL_AURA_BROKEN") then
		 spellId, spellName, spellSchool, auraType = select(9, ...);
		 if (self:IsYou(srcName, srcFlags)) then self:Output("you_dispel", srcName, srcFlags, dstName, dstFlags, spellName, "ISBROKENBY", { spellSchool = spellSchool, spellId = spellId });
		 elseif (self:IsPet(srcName, srcFlags)) then self:Output("pet_dispel", srcName, srcFlags, dstName, dstFlags, spellName, "ISBROKENBY", { spellSchool = spellSchool, spellId = spellId });
		 elseif (self:IsTarget(srcName, srcFlags)) then self:Output("mob_dispel", srcName, srcFlags, dstName, dstFlags, spellName, "ISBROKENBY", { spellSchool = spellSchool, spellId = spellId });
		 else self:Output("oth_dispel", srcName, srcFlags, dstName, dstFlags, spellName, "ISBROKENBY", { spellSchool = spellSchool, spellId = spellId });
		 end

	 elseif (eventType == "SPELL_AURA_BROKEN_SPELL") then
	  -- Source is the person breaking crowd control
	  -- Dest is the person who was being crowd controlled
	  -- Spell is the crowd control spell
	  -- ExtraSpell is the spell that broke it
		 spellId, spellName, spellSchool, extraSpellId, extraSpell, extraSpellSchool, auraType = select(9, ...);
		 extraSpell_verboseMessage = "USING";
		 if (self:IsYou(srcName, srcFlags)) then self:Output("you_break", srcName, srcFlags, dstName, dstFlags, spellName, "ISBROKENBY", { spellSchool = spellSchool, spellId = spellId, extraSpellId = extraSpellId, extraSpell = extraSpell, extraSpellSchool = extraSpellSchool, extraSpell_verboseMessage = extraSpell_verboseMessage });
		 elseif (self:IsPet(srcName, srcFlags)) then self:Output("pet_break", srcName, srcFlags, dstName, dstFlags, spellName, "ISBROKENBY", { spellSchool = spellSchool, spellId = spellId, extraSpellId = extraSpellId, extraSpell = extraSpell, extraSpellSchool = extraSpellSchool, extraSpell_verboseMessage = extraSpell_verboseMessage });
		 elseif (self:IsTarget(srcName, srcFlags)) then self:Output("mob_break", srcName, srcFlags, dstName, dstFlags, spellName, "ISBROKENBY", { spellSchool = spellSchool, spellId = spellId, extraSpellId = extraSpellId, extraSpell = extraSpell, extraSpellSchool = extraSpellSchool, extraSpell_verboseMessage = extraSpell_verboseMessage });
		 else self:Output("oth_break", srcName, srcFlags, dstName, dstFlags, spellName, "ISBROKENBY", { spellSchool = spellSchool, spellId = spellId, extraSpellId = extraSpellId, extraSpell = extraSpell, extraSpellSchool = extraSpellSchool, extraSpell_verboseMessage = extraSpell_verboseMessage });
		 end
  
	 elseif (eventType == "SPELL_INTERRUPT") then
		 -- spellName does the interrupting, extraSpellName is what gets interrupted
		 extraSpellId, extraSpell, extraSpellSchool, spellId, spellName, spellSchool = select(9, ...);
		 extraSpell_verboseMessage = "USING";
			if (self:IsYou(srcName, srcFlags)) then self:Output("you_inter", srcName, srcFlags, dstName, dstFlags, spellName, "INTERRUPT", { spellSchool = spellSchool, spellId = spellId, extraSpellId = extraSpellId, extraSpell = extraSpell, extraSpellSchool = extraSpellSchool, extraSpell_verboseMessage = extraSpell_verboseMessage });
			elseif (self:IsTarget(dstName, dstFlags)) then self:Output("mob_inter", srcName, srcFlags, dstName, dstFlags, spellName, "INTERRUPTS", { spellSchool = spellSchool, spellId = spellId, extraSpellId = extraSpellId, extraSpell = extraSpell, extraSpellSchool = extraSpellSchool, extraSpell_verboseMessage = extraSpell_verboseMessage });
			else self:Output("oth_inter", srcName, srcFlags, dstName, dstFlags, spellName, "INTERRUPTS", { spellSchool = spellSchool, spellId = spellId, extraSpellId = extraSpellId, extraSpell = extraSpell, extraSpellSchool = extraSpellSchool, extraSpell_verboseMessage = extraSpell_verboseMessage });
			end
  
  -- ENCHANT EVENTS
  		
	 elseif (eventType == "ENCHANT_APPLIED") then
		 spellName, itemId, itemName = select(9, ...);
		 if (self:IsYou(srcName, srcFlags)) then self:Output("you_ench", srcName, srcFlags, dstName, dstFlags, spellName, "APPLYENCHANTMENTON", { amount = self:GetItemLink(itemId) });
		 end
 		
	 elseif (eventType == "ENCHANT_REMOVED") then
		 spellName, itemId, itemName = select(9, ...);
		 if (self:IsYou(srcName, srcFlags)) then self:Output("you_ench", srcName, srcFlags, dstName, dstFlags, spellName, "REMOVEENCHANTMENTFROM", { amount = self:GetItemLink(itemId) });
		 end

  -- SPAM EVENTS
  
	 elseif (eventType == "SPELL_CAST_START") then
	  spellId, spellName, spellSchool = select(9, ...);
			if (self:IsYou(srcName, srcFlags)) then self:Output("spam_spellstart", srcName, srcFlags, dstName, dstFlags, spellName, "STARTCASTING", { spellSchool = spellSchool, spellId = spellId });
			else self:Output("spam_spellstart", srcName, srcFlags, dstName, dstFlags, spellName, "STARTSCASTING", { spellSchool = spellSchool, spellId = spellId });
			end

	 elseif (eventType == "SPELL_CAST_SUCCESS") then
		 spellId, spellName, spellSchool = select(9, ...);
		 srcName, dstName = self:Flip(srcName, dstName);
		 srcFlags, dstFlags = self:Flip(srcFlags, dstFlags);
			if (self:IsYou(dstName, dstFlags)) then self:Output("spam_spellsuccess", srcName, srcFlags, dstName, dstFlags, spellName, "FINISHCASTING", { spellSchool = spellSchool, spellId = spellId });
			else self:Output("spam_spellsuccess", srcName, srcFlags, dstName, dstFlags, spellName, "FINISHESCASTING", { spellSchool = spellSchool, spellId = spellId });
			end

	 elseif (eventType == "SPELL_CAST_FAILED") then
	  spellId, spellName, spellSchool, missType = select(9, ...);
			if (self:IsYou(srcName, srcFlags)) then self:Output("spam_spellfailed", srcName, srcFlags, dstName, dstFlags, spellName, "FAILTOCAST", { spellSchool = spellSchool, spellId = spellId });
			else self:Output("spam_spellfailed", srcName, srcFlags, dstName, dstFlags, spellName, "FAILSTOCAST", { spellSchool = spellSchool, spellId = spellId });
			end
 		
	 elseif (eventType == "SPELL_DISPEL_FAILED") then
	  spellId, spellName, spellSchool, extraSpellId, extraSpellName, extraSpellSchool = select(9, ...);
			if (self:IsYou(srcName, srcFlags)) then self:Output("spam_dispelfailed", srcName, srcFlags, dstName, dstFlags, extraSpellName, "FAILTODISPEL", { spellSchool = extraSpellSchool, spellId = extraSpellId });
			else self:Output("spam_dispelfailed", srcName, srcFlags, dstName, dstFlags, extraSpellName, "FAILSTODISPEL", { spellSchool = extraSpellSchool, spellId = extraSpellId });
			end

	 elseif (eventType == "SPELL_DURABILITY_DAMAGE") or (eventType == "SPELL_DURABILITY_DAMAGE_ALL") then
		 spellId, spellName, spellSchool = select(9, ...);
			if (self:IsYou(srcName, srcFlags)) then self:Output("spam_durability", srcName, srcFlags, dstName, dstFlags, spellName, "LOSEDURABILITY", { spellSchool = spellSchool, spellId = spellId });
			else self:Output("spam_durability", srcName, srcFlags, dstName, dstFlags, spellName, "LOSESDURABILITY", { spellSchool = spellSchool, spellId = spellId });
			end

	 elseif (eventType == "SPELL_INSTAKILL") then
		 spellId, spellName, spellSchool = select(9, ...);
			if (self:IsYou(srcName, srcFlags)) then self:Output("spam_instantkill", srcName, srcFlags, dstName, dstFlags, spellName, "INSTANTLYKILL", { spellSchool = spellSchool, spellId = spellId });
			else self:Output("spam_instantkill", srcName, srcFlags, dstName, dstFlags, spellName, "INSTANTLYKILLS", { spellSchool = spellSchool, spellId = spellId });
			end

	 elseif (eventType == "UNIT_DIED") or (eventType == "UNIT_DISSIPATES") then
	  if (self:IsYou(dstName, dstFlags)) then self:Output("mob_died", dstName, dstFlags, nil, nil, nil, "DIE");
	  else self:Output("spam_unitdied", srcName, srcFlags, dstName, dstFlags, nil, "DIES");
	  end
			
	 elseif (eventType == "UNIT_DESTROYED") then
	  -- for example, summoned elemental ending its allotted life
			self:Output("spam_unitdestroyed", srcName, srcFlags, dstName, dstFlags, nil, "ISDESTROYED");

  -- UNKNOWN EVENT
  
  else
  
  end
  
  
 -- COMBAT STARTED ------------------------------------------------------------------------------------------------------------
 
 elseif (event == "PLAYER_REGEN_DISABLED") then
  inCombat = 1;
  damageTaken = 0;
  damageGiven = 0;
  healsTaken = 0;
  healsGiven = 0;
  damageResisted = 0;
  petDamageGiven = 0;
  petDamageTaken = 0;
		petHealsTaken = 0;
		petHealsGiven = 0;
  manaGained = 0;
  energyGained = 0;
  rageGained = 0;
  attacksGained = 0;
  skillsIncreased = 0;
  expGained = 0;
  combatStartTime = GetTime();
  announced = false;
 
 
 -- COMBAT ENDED --------------------------------------------------------------------------------------------------------------
 
 elseif (event == "PLAYER_REGEN_ENABLED") then
  self:SummarizeCombat();
  inCombat = 0;
  damageTaken = 0;
  damageGiven = 0;
  healsTaken = 0;
  healsGiven = 0;
  damageResisted = 0;
  petDamageGiven = 0;
  petDamageTaken = 0;
		petHealsTaken = 0;
		petHealsGiven = 0;
  manaGained = 0;
  energyGained = 0;
  rageGained = 0;
  attacksGained = 0;
  skillsIncreased = 0;
  expGained = 0;
  
  
 -- REPUTATION ----------------------------------------------------------------------------------------------------------------
 elseif (event == "CHAT_MSG_COMBAT_FACTION_CHANGE") then
 
  for spell, damage in string.gmatch(arg1, self.ChatPatterns[1]) do -- Reputation with (.+) increased by (%d+).
			if (self:Get("flagexp") == 1) then damage = "+"..damage; end
			self:Output("you_gain_rep", UnitName("player"), HM_COMBATLOG_OBJECT_YOU, nil, nil, nil, "GAINREP", { amount = spell.." "..damage, expType = HM_EXPTYPE_REPUTATION });
   return;
  end
  for spell, damage in string.gmatch(arg1, self.ChatPatterns[2]) do -- Reputation with (.+) decreased by (%d+).
			if (self:Get("flagexp") == 1) then damage = "-"..damage; end
			self:Output("you_gain_rep", UnitName("player"), HM_COMBATLOG_OBJECT_YOU, nil, nil, nil, "LOSEREP", { amount = spell.." "..damage, expType = HM_EXPTYPE_REPUTATION });
   return;
  end
  

 -- LOOT ----------------------------------------------------------------------------------------------------------------------

 elseif (event == "CHAT_MSG_LOOT" or event == "CHAT_MSG_MONEY") then
 
		self:DebugWrite(string.format("Parser received loot event: %s", event));
 
  for spell in string.gmatch(arg1, self.ChatPatterns[4]) do -- You receive loot: (.+).
			self:Output("you_loot", UnitName("player"), HM_COMBATLOG_OBJECT_YOU, nil, nil, nil, "PLAYERLOOT", { amount = spell });
   return;
  end
  for spell in string.gmatch(arg1, self.ChatPatterns[12]) do -- You loot (.+)
			self:Output("you_loot", UnitName("player"), HM_COMBATLOG_OBJECT_YOU, nil, nil, nil, "PLAYERLOOT", { amount = spell });
   return;
  end
  for spell in string.gmatch(arg1, self.ChatPatterns[5]) do -- Your share of the loot is (.+).
			self:Output("you_loot", UnitName("player"), HM_COMBATLOG_OBJECT_YOU, nil, nil, nil, "PLAYERLOOT", { amount = spell });
   return;
  end
  for name, spell in string.gmatch(arg1, self.ChatPatterns[6]) do -- (.+) receives loot: (.+).
			self:Output("they_loot", name, HM_COMBATLOG_OBJECT_PARTYMEMBER, nil, nil, nil, "PLAYERLOOTS", { amount = spell });
   return;
  end
  for spell in string.gmatch(arg1, self.ChatPatterns[3]) do -- You create: (.+).
			self:Output("you_craft", UnitName("player"), HM_COMBATLOG_OBJECT_YOU, nil, nil, nil, "PLAYERCRAFT", { amount = spell });
   return;
  end
 
 
 -- SKILL ---------------------------------------------------------------------------------------------------------------------

 elseif (event == "CHAT_MSG_SKILL") then

  for spell, damage in string.gmatch(arg1, self.ChatPatterns[7]) do -- Your skill in (.+) has increased to ([%d]+)
			self:Output("skill_up", UnitName("player"), HM_COMBATLOG_OBJECT_YOU, nil, nil, spell, "SKILLHASINCREASEDTO", { amount = damage });
   skillsIncreased = self:Add(skillsIncreased, 1);
   return;
  end
 

 -- HONOR ---------------------------------------------------------------------------------------------------------------------

 elseif (event == "CHAT_MSG_COMBAT_HONOR_GAIN") then

  for spell in string.gmatch(arg1, self.ChatPatterns[8]) do -- ([%d]+)
			if (self:Get("flagexp") == 1) then spell = "+"..spell; end
			self:Output("you_gain_honor", UnitName("player"), HM_COMBATLOG_OBJECT_YOU, nil, nil, nil, "HONOR", { amount = spell, expType = HM_EXPTYPE_HONOR });
   return;
  end
  for spell in string.gmatch(arg1, self.ChatPatterns[9]) do -- You have been awarded ([%d]+) honor.
			if (self:Get("flagexp") == 1) then spell = "+"..spell; end
			self:Output("you_gain_honor", UnitName("player"), HM_COMBATLOG_OBJECT_YOU, nil, nil, nil, "HONOR", { amount = spell, expType = HM_EXPTYPE_HONOR });
   return;
  end


 -- EXPERIENCE ----------------------------------------------------------------------------------------------------------------

 elseif (event == "CHAT_MSG_COMBAT_XP_GAIN") then
  _, _, d = string.find(arg1, self.ChatPatterns[10]); -- .* gain (%d+).*
  _, _, d2 = string.find(arg1, self.ChatPatterns[11]); -- experience%. %(%+(%d+) exp Rested bonus
  expGained = self:Add(expGained, d);
  if (d2 ~= nil and self:Get("showrested") == 1) then d = (d or "").." (+"..d2..")"; end
  if (self:Get("flagexp") == 1) then d = "+"..(d or ""); end
		self:Output("exp", UnitName("player"), HM_COMBATLOG_OBJECT_YOU, nil, nil, nil, "XP", { amount = (d or 0), expType = HM_EXPTYPE_EXPERIENCE });


 ------------------------------------------------------------------------------------------------------------------------------
 
 else
  self:Print("Unknown event encountered! > "..self:FormatCombatEvent(...));


	end
	
end


----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

--
-- All damage done is routed through this handler.
--
function HitsMode:OutputDamage(eventType, srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, amount, school, resisted, blocked, absorbed, critical, glancing, crushing, environmentalType, overkill)
 
	self:DebugWrite(string.format(GRN.."OutputDamage received eventType: %s || src: %s || dst: %s || amount: %s|r", eventType or "nil", srcName or "nil", dstName or "nil", amount or "nil"));
	if (amount == nil) then return; end
	
 local option, verboseMessage;
	local hitsfor, afflicts;
	
	if (self:IsCrit(critical)) then
		hitsfor = "CRITSFOR";
		afflicts = "CRITSFOR";
	else
		hitsfor = "HITSFOR";
		afflicts = "AFFLICTS";
	end

 verboseMessage = hitsfor;

 -- ACTIONS TAKEN BY YOU -----------------------
 if (self:IsYou(srcName, srcFlags)) then
 
		damageGiven = self:Add(damageGiven, amount);
		
		if (eventType == "SWING_DAMAGE" or eventType == "RANGE_DAMAGE" or eventType == "DAMAGE_SHIELD" or eventType == "DAMAGE_SPLIT" or eventType == "SPELL_BUILDING_DAMAGE") then
			option = "you_melee";
			
		elseif (eventType == "SPELL_DAMAGE") then
		 option = "you_nuke";
		
		elseif (eventType == "SPELL_PERIODIC_DAMAGE") then
		 verboseMessage = afflicts;
		 option = "you_dot";
					
		end
		 
 -- ACTIONS TAKEN BY YOUR PET ------------------
 elseif (self:IsPet(srcName, srcFlags)) then
 
		petDamageGiven = self:Add(petDamageGiven, amount);
		
		if (eventType == "SWING_DAMAGE" or eventType == "RANGE_DAMAGE" or eventType == "DAMAGE_SHIELD" or eventType == "DAMAGE_SPLIT" or eventType == "SPELL_BUILDING_DAMAGE") then
		 option = "pet_melee";

		elseif (eventType == "SPELL_DAMAGE") then
		 option = "pet_nuke";
		
		elseif (eventType == "SPELL_PERIODIC_DAMAGE") then
		 verboseMessage = afflicts;
		 option = "pet_dot";

		end
		
 -- ATTACKER VERSUS YOU OR ITSELF --------------
 elseif (self:IsYou(dstName, dstFlags)) or (self:IsTarget(srcName, srcFlags) and self:IsTarget(dstName, dstFlags)) then

		damageTaken = self:Add(damageTaken, amount);
		
		if (eventType == "SWING_DAMAGE" or eventType == "RANGE_DAMAGE" or eventType == "DAMAGE_SHIELD" or eventType == "DAMAGE_SPLIT" or eventType == "SPELL_BUILDING_DAMAGE") then
		 option = "mob_melee";

		elseif (eventType == "SPELL_DAMAGE") then
		 option = "mob_nuke";
		
		elseif (eventType == "SPELL_PERIODIC_DAMAGE") then
		 verboseMessage = afflicts;
		 option = "mob_dot";

		elseif (eventType == "ENVIRONMENTAL_DAMAGE") then
		 option = "you_fall";
			if (environmentalType == "DROWNING") then verboseMessage = "DROWNFOR";
			elseif (environmentalType == "FALLING") then verboseMessage = "FALLFOR";
			elseif (environmentalType == "FATIGUE") then verboseMessage = "AREFATIGUEDFOR";
			elseif (environmentalType == "FIRE") then verboseMessage = "BURNFOR";
			elseif (environmentalType == "LAVA") then verboseMessage = "BURNFOR";
			elseif (environmentalType == "SLIME") then verboseMessage = "ARESLIMEDFOR";
			end

		end
		
 -- ATTACKER VERSUS YOUR PET -------------------
 elseif (self:IsPet(dstName, dstFlags)) then
 
		petDamageTaken = self:Add(petDamageTaken, amount);
		
		if (eventType == "SWING_DAMAGE" or eventType == "RANGE_DAMAGE" or eventType == "DAMAGE_SHIELD" or eventType == "DAMAGE_SPLIT" or eventType == "SPELL_BUILDING_DAMAGE") then
		 option = "vspet_melee";

		elseif (eventType == "SPELL_DAMAGE") then
		 option = "vspet_nuke";
		
		elseif (eventType == "SPELL_PERIODIC_DAMAGE") then
		 verboseMessage = afflicts;
		 option = "vspet_dot";

		end

 -- OTHERS VERSUS OTHERS -----------------------
	else

		if (eventType == "SWING_DAMAGE" or eventType == "RANGE_DAMAGE" or eventType == "DAMAGE_SHIELD" or eventType == "DAMAGE_SPLIT" or eventType == "SPELL_BUILDING_DAMAGE") then
		 option = "oth_melee";

		elseif (eventType == "SPELL_DAMAGE") then
		 option = "oth_nuke";
		
		elseif (eventType == "SPELL_PERIODIC_DAMAGE") then
		 verboseMessage = afflicts;
		 option = "oth_dot";

		elseif (eventType == "ENVIRONMENTAL_DAMAGE") then
		 option = "oth_fall";
			if (environmentalType == "DROWNING") then verboseMessage = "DROWNSFOR";
			elseif (environmentalType == "FALLING") then verboseMessage = "FALLSFOR";
			elseif (environmentalType == "FATIGUE") then verboseMessage = "ISFATIGUEDFOR";
			elseif (environmentalType == "FIRE") then verboseMessage = "BURNSFOR";
			elseif (environmentalType == "LAVA") then verboseMessage = "BURNSFOR";
			elseif (environmentalType == "SLIME") then verboseMessage = "ISSLIMEDFOR";
			end

		end
		
 end
 
 -- Output
	self:Output(option, srcName, srcFlags, dstName, dstFlags, spellName, verboseMessage, {
	 amount = self:Damage(amount, critical),
	 critical = self:IsCrit(critical),
	 spellSchool = spellSchool,
	 resisted = resisted,
	 blocked = blocked,
	 absorbed = absorbed,
	 crushing = crushing,
	 glancing = glancing, 
	 spellId = spellId,
	 overkill = overkill
	});

end


--
-- All misses are routed through this handler.
--
function HitsMode:OutputMiss(eventType, srcName, srcFlags, dstName, dstFlags, missType, spellId, spellName, spellSchool, amountMissed)
 
	self:DebugWrite(string.format(GRN.."OutputMiss received eventType: %s || src: %s || dst: %s || missType: %s || spellName: %s|r", eventType or "nil", srcName or "nil", dstName or "nil", missType or "nil", spellName or "nil"));
	if (missType == nil) then return; end
	
	local amount = nil;
	if (missType == "RESIST" or missType == "BLOCK" or missType == "ABSORB") then
	 if (amountMissed or 0 ~= 0) then amount = amountMissed; end
	end

 local option, verboseMessage;
 
 if (self:IsYou(dstName, dstFlags)) or (self:IsTarget(srcName, srcFlags) and self:IsTarget(dstName, dstFlags)) then
  if (missType == "MISS") then verboseMessage = "MISSES";
		elseif (missType == "DODGE") then verboseMessage = "DODGE";
		elseif (missType == "ABSORB") then verboseMessage = "ABSORB";
		elseif (missType == "RESIST") then verboseMessage = "RESIST";
		elseif (missType == "PARRY") then verboseMessage = "PARRY";
		elseif (missType == "BLOCK") then verboseMessage = "BLOCK";
		elseif (missType == "REFLECT") then verboseMessage = "REFLECT";
		elseif (missType == "DEFLECT") then verboseMessage = "DEFLECT";
		elseif (missType == "IMMUNE") then verboseMessage = "AREIMMUNE";
		elseif (missType == "EVADE") then verboseMessage = "EVADE";
		end
	else
  if (missType == "MISS") then verboseMessage = "MISS";
		elseif (missType == "DODGE") then verboseMessage = "DODGES";
		elseif (missType == "ABSORB") then verboseMessage = "ABSORBS";
		elseif (missType == "RESIST") then verboseMessage = "RESISTS";
		elseif (missType == "PARRY") then verboseMessage = "PARRIES";
		elseif (missType == "BLOCK") then verboseMessage = "BLOCKS";
		elseif (missType == "REFLECT") then verboseMessage = "REFLECTS";
		elseif (missType == "DEFLECT") then verboseMessage = "DEFLECTS";
		elseif (missType == "IMMUNE") then verboseMessage = "ISIMMUNE";
		elseif (missType == "EVADE") then verboseMessage = "EVADES";
		end
	end

 -- ACTIONS TAKEN BY YOU -----------------------
 if (self:IsYou(srcName, srcFlags)) then
  if (missType == "MISS") then option = "you_missmiss"; else option = "mob_miss"; end
		 
 -- ACTIONS TAKEN BY YOUR PET ------------------
 elseif (self:IsPet(srcName, srcFlags)) then
  if (missType == "MISS") then option = "pet_missmiss"; else option = "vspet_miss"; end
		
 -- ATTACKER VERSUS YOU OR ITSELF --------------
 elseif (self:IsYou(dstName, dstFlags)) or (self:IsTarget(srcName, srcFlags) and self:IsTarget(dstName, dstFlags)) then
  if (missType == "MISS") then option = "mob_missmiss"; else option = "you_miss"; end
		
 -- ATTACKER VERSUS YOUR PET -------------------
 elseif (self:IsPet(dstName, dstFlags)) then
  if (missType == "MISS") then option = "mob_missmiss"; else option = "pet_miss"; end

 -- OTHERS VERSUS OTHERS -----------------------
	else
  if (missType == "MISS") then option = "oth_missmiss"; else option = "oth_miss"; end
		
 end

 -- Output
	self:Output(option, srcName, srcFlags, dstName, dstFlags, spellName, verboseMessage, {
	 amount = amount,
	 spellSchool = spellSchool,
	 missType = missType,
	 spellId = spellId
	});
	
end


--
-- All heals are routed through this handler.
--
function HitsMode:OutputHeal(eventType, srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, amount, critical, overhealing, absorbed)
 
	self:DebugWrite(string.format(RED.."OutputHeal received eventType: %s || src: %s || dst: %s || spellName: %s || amount: %s || crit: %s|r", eventType or "nil", srcName or "nil", dstName or "nil", spellName or "nil", amount or "nil", critical or "nil"));
	if (amount == nil) then return; end

 local option, extraOption, verboseMessage;
 
 verboseMessage = "HEALSFOR";
 extraOption = nil;

 -- YOU HEAL OTHERS ----------------------------
 if (self:IsYou(srcName, srcFlags) and self:IsYou(dstName, dstFlags) == false and self:IsPet(dstName, dstFlags) == false) then

		healsGiven = self:Add(healsGiven, amount);

		if (eventType == "SPELL_HEAL" or eventType == "SPELL_BUILDING_HEAL") then option = "you_heal";
		elseif (eventType == "SPELL_PERIODIC_HEAL") then option = "you_hot_others";
		end

 -- YOU HEAL YOURSELF --------------------------
 elseif (self:IsYou(srcName, srcFlags) and self:IsYou(dstName, dstFlags)) then

		-- This qualifies as both a heal given and received
		healsGiven = self:Add(healsGiven, amount);
		healsTaken = self:Add(healsTaken, amount);
		
		-- Fire two events here in case heals given and heals received are being routed to different windows, they can see events in both windows
		if (eventType == "SPELL_HEAL" or eventType == "SPELL_BUILDING_HEAL") then
		 option = "you_heal";
		 extraOption = "you_heal_yourself";
		elseif (eventType == "SPELL_PERIODIC_HEAL") then
		 option = "you_hot_others";
		 extraOption = "you_hot_yourself";
		end
		 
 -- YOU HEAL YOUR PET --------------------------
 elseif (self:IsYou(srcName, srcFlags) and self:IsPet(dstName, dstFlags)) then

		-- This qualifies as both a heal given and received
		healsGiven = self:Add(healsGiven, amount);
		petHealsTaken = self:Add(petHealsTaken, amount);

		-- Fire two events here in case heals given and heals received are being routed to different windows, they can see events in both windows
		if (eventType == "SPELL_HEAL" or eventType == "SPELL_BUILDING_HEAL") then
		 option = "you_heal";
		 extraOption = "you_heal_pet";
		elseif (eventType == "SPELL_PERIODIC_HEAL") then
		 option = "you_hot_others";
		 extraOption = "you_hot_pet";
		end

 -- SOMEONE ELSE HEALS YOU ---------------------
 elseif (self:IsYou(dstName, dstFlags)) then

		healsTaken = self:Add(healsTaken, amount);

		if (eventType == "SPELL_HEAL" or eventType == "SPELL_BUILDING_HEAL") then option = "others_heal_you";
		elseif (eventType == "SPELL_PERIODIC_HEAL") then option = "others_hot_you";
		end
		
 -- SOMEONE ELSE HEALS YOUR PET ----------------
 elseif (self:IsPet(dstName, dstFlags)) then

		petHealsTaken = self:Add(petHealsTaken, amount);

		if (eventType == "SPELL_HEAL" or eventType == "SPELL_BUILDING_HEAL") then option = "others_heal_pet";
		elseif (eventType == "SPELL_PERIODIC_HEAL") then option = "others_hot_pet";
		end
		
 -- YOUR PET HEALS SOMEONE ---------------------
 elseif (self:IsPet(srcName, srcFlags)) then

		healsGiven = self:Add(healsGiven, amount);

		if (eventType == "SPELL_HEAL" or eventType == "SPELL_BUILDING_HEAL") then option = "you_heal";
		elseif (eventType == "SPELL_PERIODIC_HEAL") then option = "you_hot_others";
		end
		
 -- YOUR TARGET HEALS ITSELF -------------------
 elseif (self:IsTarget(srcName, srcFlags) and self:IsTarget(dstName, dstFlags)) then
 
		if (eventType == "SPELL_HEAL" or eventType == "SPELL_BUILDING_HEAL") then option = "mob_heal_itself";
		elseif (eventType == "SPELL_PERIODIC_HEAL") then option = "mob_hot_itself";
		end
		
 -- SOMEONE ELSE HEALS SOMEONE ELSE ------------
	else

		if (eventType == "SPELL_HEAL" or eventType == "SPELL_BUILDING_HEAL") then option = "oth_heal_oth";
		elseif (eventType == "SPELL_PERIODIC_HEAL") then option = "oth_hot_oth";
		end
		
 end
 
 -- Output
 self:Output(option, srcName, srcFlags, dstName, dstFlags, spellName, verboseMessage, {
  amount = amount,
  critical = self:IsCrit(critical),
  spellSchool = spellSchool,
  spellId = spellId,
  overhealing = overhealing,
	 absorbed = absorbed
 });
 if (extraOption ~= nil) then self:Output(extraOption, srcName, srcFlags, dstName, dstFlags, spellName, verboseMessage, {
  amount = amount,
  critical = self:IsCrit(critical),
  spellSchool = spellSchool,
  spellSchool = spellId,
  overhealing = overhealing,
	 absorbed = absorbed
 }); end

end


--
-- All energizes are routed through this handler.
--
function HitsMode:OutputEnergize(eventType, srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, amount, powerType)
 
	self:DebugWrite(string.format(YEL.."OutputEnergize received eventType: %s || src: %s || dst: %s || powerType: %s || spellName: %s|r", eventType or "nil", srcName or "nil", dstName or "nil", powerType or "nil", spellName or "nil"));
	if (powerType == nil) then return; end
	
	if (
	 powerType ~= -2 and
	 powerType ~= SPELL_POWER_MANA and
	 powerType ~= SPELL_POWER_RAGE and
	 powerType ~= SPELL_POWER_FOCUS and
	 powerType ~= SPELL_POWER_ENERGY and
	 powerType ~= SPELL_POWER_HAPPINESS and
	 powerType ~= SPELL_POWER_RUNES and
	 powerType ~= SPELL_POWER_RUNIC_POWER) then
	  self:Print("Unknown Power Type (Energize): "..powerType);
	end

	-- Reroute health gains to the heal handler (hopefully this works)
	if (powerType == -2) then
		self:OutputHeal("SPELL_PERIODIC_HEAL", srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, amount, nil);
		return;
	end
	
	local option, verboseMessage;
	
	-- Remove the destination if it's the same as the source
	--if (srcName == dstName) then
	 --dstName = nil;
	 --dstFlags = nil;
	--end
	
 if (self:IsYou(dstName, dstFlags) or (self:IsYou(srcName, srcFlags) and dstName == nil)) then
		if ((spellName or "") ~= "") then
		 if (powerType == SPELL_POWER_MANA) then verboseMessage = "GAINMANAFROM";
		 elseif (powerType == SPELL_POWER_RAGE) then verboseMessage = "GAINRAGEFROM";
		 elseif (powerType == SPELL_POWER_FOCUS) then verboseMessage = "GAINFOCUSFROM";
		 elseif (powerType == SPELL_POWER_ENERGY) then verboseMessage = "GAINENERGYFROM";
		 elseif (powerType == SPELL_POWER_RUNES) then verboseMessage = "GAINRUNESFROM";
		 elseif (powerType == SPELL_POWER_RUNIC_POWER) then verboseMessage = "GAINRUNICPOWERFROM";
		 end
		else
		 if (powerType == SPELL_POWER_MANA) then verboseMessage = "GAINMANA";
		 elseif (powerType == SPELL_POWER_RAGE) then verboseMessage = "GAINRAGE";
		 elseif (powerType == SPELL_POWER_FOCUS) then verboseMessage = "GAINFOCUS";
		 elseif (powerType == SPELL_POWER_ENERGY) then verboseMessage = "GAINENERGY";
		 elseif (powerType == SPELL_POWER_RUNES) then verboseMessage = "GAINRUNES";
		 elseif (powerType == SPELL_POWER_RUNIC_POWER) then verboseMessage = "GAINRUNICPOWER";
		 end
		end
 else
		if (powerType == SPELL_POWER_MANA) then verboseMessage = "GAINSMANAFROM";
		elseif (powerType == SPELL_POWER_RAGE) then verboseMessage = "GAINSRAGEFROM";
		elseif (powerType == SPELL_POWER_FOCUS) then verboseMessage = "GAINSFOCUSFROM";
		elseif (powerType == SPELL_POWER_ENERGY) then verboseMessage = "GAINSENERGYFROM";
		elseif (powerType == SPELL_POWER_HAPPINESS) then verboseMessage = "INCREASESHAPPINESSFOR";
		elseif (powerType == SPELL_POWER_RUNES) then verboseMessage = "GAINSRUNESFROM";
		elseif (powerType == SPELL_POWER_RUNIC_POWER) then verboseMessage = "GAINSRUNICPOWERFROM";
		end
 end
	
 -- YOU ARE ENERGIZED --------------------------
 if (self:IsYou(dstName, dstFlags) or (self:IsYou(srcName, srcFlags) and dstName == nil)) then
 
		-- Accumulate DPS. We don't track focus or pet happiness.
		if (powerType == SPELL_POWER_MANA) then manaGained = self:Add(manaGained, amount);
		elseif (powerType == SPELL_POWER_RAGE) then rageGained = self:Add(rageGained, amount);
		elseif (powerType == SPELL_POWER_ENERGY) then energyGained = self:Add(energyGained, amount);
		end
		
		option = "you_energize";
		 
 -- YOUR PET IS ENERGIZED ----------------------
 elseif (self:IsPet(dstName, dstFlags)) then
  option = "pet_energize";
		
 -- YOUR TARGET IS ENERGIZED -------------------
 elseif (self:IsTarget(dstName, dstFlags)) then
  option = "mob_energize";
		
 -- SOMEONE ELSE IS ENERGIZED ------------------
	else
	 option = "oth_energize";
		
 end

 --Output
	self:Output(option, srcName, srcFlags, dstName, dstFlags, spellName, verboseMessage, {
	 amount = amount,
	 spellSchool = spellSchool,
	 spellId = spellId,
	 powerType = powerType
	});

end


--
-- All aura events are routed through this handler.
--
function HitsMode:OutputAura(eventType, srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, auraType, amount)

	self:DebugWrite2(string.format(YEL.."OutputAura received eventType: %s || src: %s || dst: %s || auraType: %s || spellName: %s || amount: %s|r", eventType or "nil", srcName or "nil", dstName or "nil", auraType or "nil", spellName or "nil", amount or "nil"));
 if ((auraType or "") == "") then return; end

 local option, verboseMessage, auraChangeType, showOn;
 
 -- Set the aura change type for the output engine
 if (eventType == "SPELL_AURA_APPLIED") or (eventType == "SPELL_AURA_APPLIED_DOSE") then auraChangeType = "GAIN";
 elseif (eventType == "SPELL_AURA_REMOVED") or (eventType == "SPELL_AURA_REMOVED_DOSE") then auraChangeType = "FADE";
 elseif (eventType == "SPELL_AURA_REFRESH") then auraChangeType = "REFRESH";
 end
 
 if ((srcName or "") == "") then
 
  -- AN AURA WAS GIVEN TO YOU WITH NO SOURCE ------------------
  if (self:IsYou(dstName, dstFlags)) then
   srcName = dstName;
   srcFlags = dstFlags;
   dstName = nil;
   dstFlags = nil;
   if (eventType == "SPELL_AURA_APPLIED") or (eventType == "SPELL_AURA_APPLIED_DOSE") then
    option = "you_gain";
    if (auraType == "BUFF") then verboseMessage = "GAIN"; else verboseMessage = "AREAFFLICTED"; end
   elseif (eventType == "SPELL_AURA_REMOVED") or (eventType == "SPELL_AURA_REMOVED_DOSE") then
    option = "you_debuff";
    verboseMessage = "FADES";
   elseif (eventType == "SPELL_AURA_REFRESH") then
    if (auraType == "BUFF") then option = "you_refresh"; else option = "you_refresh_debuff"; end
    verboseMessage = "ISREFRESHED";
   end
  
  -- AN AURA WAS GIVEN TO YOUR PET WITH NO SOURCE -------------
  elseif (self:IsPet(dstName, dstFlags)) then
   srcName = dstName;
   srcFlags = dstFlags;
   dstName = nil;
   dstFlags = nil;
   if (eventType == "SPELL_AURA_APPLIED") or (eventType == "SPELL_AURA_APPLIED_DOSE") then
    option = "pet_gains";
    if (auraType == "BUFF") then verboseMessage = "GAINS"; else verboseMessage = "ISAFFLICTED"; end
   elseif (eventType == "SPELL_AURA_REMOVED") or (eventType == "SPELL_AURA_REMOVED_DOSE") then
    option = "pet_debuff";
    verboseMessage = "FADES";
   elseif (eventType == "SPELL_AURA_REFRESH") then
    if (auraType == "BUFF") then option = "pet_refresh"; else option = "pet_refresh_debuff"; end
    verboseMessage = "ISREFRESHED";
   end

  -- AN AURA WAS GIVEN TO YOUR TARGET WITH NO SOURCE ----------
  elseif (self:IsTarget(dstName, dstFlags)) then
   srcName = dstName;
   srcFlags = dstFlags;
   dstName = nil;
   dstFlags = nil;
   if (eventType == "SPELL_AURA_APPLIED") or (eventType == "SPELL_AURA_APPLIED_DOSE") then
    option = "mob_gains";
    if (auraType == "BUFF") then verboseMessage = "GAINS"; else verboseMessage = "ISAFFLICTED"; end
   elseif (eventType == "SPELL_AURA_REMOVED") or (eventType == "SPELL_AURA_REMOVED_DOSE") then
    option = "mob_debuff";
    verboseMessage = "FADES";
   elseif (eventType == "SPELL_AURA_REFRESH") then
    if (auraType == "BUFF") then option = "mob_refresh"; else option = "mob_refresh_debuff"; end
    verboseMessage = "ISREFRESHED";
   end

  -- AN AURA WAS GIVEN TO SOMEONE ELSE WITH NO SOURCE ---------
  else
   srcName = dstName;
   srcFlags = dstFlags;
   dstName = nil;
   dstFlags = nil;
   if (eventType == "SPELL_AURA_APPLIED") or (eventType == "SPELL_AURA_APPLIED_DOSE") then
    option = "oth_gain";
    if (auraType == "BUFF") then verboseMessage = "GAINS"; else verboseMessage = "ISAFFLICTED"; end
   elseif (eventType == "SPELL_AURA_REMOVED") or (eventType == "SPELL_AURA_REMOVED_DOSE") then
    option = "oth_debuff";
    verboseMessage = "FADES";
   elseif (eventType == "SPELL_AURA_REFRESH") then
    if (auraType == "BUFF") then option = "oth_refresh"; else option = "oth_refresh_debuff"; end
    verboseMessage = "ISREFRESHED";
   end
  end
  
 else
 
  -- YOU GIVE AN AURA TO ANYONE -------------------------------
  if (self:IsYou(srcName, srcFlags)) then
   if (eventType == "SPELL_AURA_APPLIED") or (eventType == "SPELL_AURA_APPLIED_DOSE") then
    option = "you_gain";
    if (auraType == "BUFF") then verboseMessage = "CAST"; else verboseMessage = "CAST"; end
    showOn = true;
   elseif (eventType == "SPELL_AURA_REMOVED") or (eventType == "SPELL_AURA_REMOVED_DOSE") then
    option = "you_debuff";
    if (auraType == "BUFF") then verboseMessage = "FADESFROM"; else verboseMessage = "FADESFROM"; end
   elseif (eventType == "SPELL_AURA_REFRESH") then
    if (auraType == "BUFF") then option = "you_refresh"; else option = "you_refresh_debuff"; end
    verboseMessage = "ISREFRESHED";
   end
   
  -- YOUR PET GIVES AN AURA TO ANYONE -------------------------
  elseif (self:IsPet(srcName, srcFlags)) then
   if (eventType == "SPELL_AURA_APPLIED") or (eventType == "SPELL_AURA_APPLIED_DOSE") then
    option = "pet_gains";
    if (auraType == "BUFF") then verboseMessage = "CASTS"; else verboseMessage = "CASTS"; end
    showOn = true;
   elseif (eventType == "SPELL_AURA_REMOVED") or (eventType == "SPELL_AURA_REMOVED_DOSE") then
    option = "pet_debuff";
    if (auraType == "BUFF") then verboseMessage = "FADESFROM"; else verboseMessage = "FADESFROM"; end
   elseif (eventType == "SPELL_AURA_REFRESH") then
    if (auraType == "BUFF") then option = "pet_refresh"; else option = "pet_refresh_debuff"; end
    verboseMessage = "ISREFRESHED";
   end
   
  -- SOMEONE GIVES AN AURA TO YOU -----------------------------
  elseif (self:IsYou(dstName, dstFlags)) then
   if (eventType == "SPELL_AURA_APPLIED") or (eventType == "SPELL_AURA_APPLIED_DOSE") then
    option = "mob_gains";
    if (auraType == "BUFF") then verboseMessage = "CASTS"; else verboseMessage = "CASTS"; end
    showOn = true;
   elseif (eventType == "SPELL_AURA_REMOVED") or (eventType == "SPELL_AURA_REMOVED_DOSE") then
    option = "mob_debuff";
    if (auraType == "BUFF") then verboseMessage = "FADESFROM"; else verboseMessage = "FADESFROM"; end
   elseif (eventType == "SPELL_AURA_REFRESH") then
    if (auraType == "BUFF") then option = "mob_refresh"; else option = "mob_refresh_debuff"; end
    verboseMessage = "ISREFRESHED";
   end
   
  -- SOMEONE GIVES AN AURA TO YOUR PET ------------------------
  elseif (self:IsPet(dstName, dstFlags)) then
   if (eventType == "SPELL_AURA_APPLIED") or (eventType == "SPELL_AURA_APPLIED_DOSE") then
    option = "vspet_gains";
    if (auraType == "BUFF") then verboseMessage = "CASTS"; else verboseMessage = "CASTS"; end
    showOn = true;
   elseif (eventType == "SPELL_AURA_REMOVED") or (eventType == "SPELL_AURA_REMOVED_DOSE") then
    option = "vspet_debuff";
    if (auraType == "BUFF") then verboseMessage = "FADESFROM"; else verboseMessage = "FADESFROM"; end
   elseif (eventType == "SPELL_AURA_REFRESH") then
    if (auraType == "BUFF") then option = "vspet_refresh"; else option = "vspet_refresh_debuff"; end
    verboseMessage = "ISREFRESHED";
   end
   
  -- ANYTHING ELSE --------------------------------------------
  else
   if (eventType == "SPELL_AURA_APPLIED") or (eventType == "SPELL_AURA_APPLIED_DOSE") then
    option = "oth_gain";
    if (auraType == "BUFF") then verboseMessage = "CASTS"; else verboseMessage = "CASTS"; end
    showOn = true;
   elseif (eventType == "SPELL_AURA_REMOVED") or (eventType == "SPELL_AURA_REMOVED_DOSE") then
    option = "oth_debuff";
    if (auraType == "BUFF") then verboseMessage = "FADESFROM"; else verboseMessage = "FADESFROM"; end
   elseif (eventType == "SPELL_AURA_REFRESH") then
    if (auraType == "BUFF") then option = "oth_refresh"; else option = "oth_refresh_debuff"; end
    verboseMessage = "ISREFRESHED";
   end
   
  end
  
 end

 -- Substitute "diminishes" for "fades" if needed
 if (eventType == "SPELL_AURA_REMOVED_DOSE" and (amount or 0) > 0) then verboseMessage = "DIMINISHES"; end

 -- Output
 self:Output(option, srcName, srcFlags, dstName, dstFlags, spellName, verboseMessage, {
  amount = amount,
  spellSchool = spellSchool,
  spellId = spellId,
  auraType = auraType,
  auraChangeType = auraChangeType,
  showOn = showOn
 });

end


--
-- All drains are routed through this handler.
--
function HitsMode:OutputDrain(eventType, srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, amount, powerType, extraAmount)
	
	self:DebugWrite(string.format(YEL.."OutputDrain received eventType: %s || src: %s || dst: %s || powerType: %s || spellName: %s|r", eventType or "nil", srcName or "nil", dstName or "nil", powerType or "nil", spellName or "nil"));
	if (powerType == nil) then return; end
	
	if (
	 powerType ~= -2 and
	 powerType ~= SPELL_POWER_MANA and
	 powerType ~= SPELL_POWER_RAGE and
	 powerType ~= SPELL_POWER_FOCUS and
	 powerType ~= SPELL_POWER_ENERGY and
	 powerType ~= SPELL_POWER_HAPPINESS and
	 powerType ~= SPELL_POWER_RUNES and
	 powerType ~= SPELL_POWER_RUNIC_POWER) then
	  self:Print("Unknown Power Type (Drain) > "..powerType);
	end
	
	local option, verboseMessage;
	
	if (powerType == SPELL_POWER_MANA) then verboseMessage = "DRAINSMANA";
	elseif (powerType == SPELL_POWER_RAGE) then verboseMessage = "DRAINSRAGE";
	elseif (powerType == SPELL_POWER_FOCUS) then verboseMessage = "DRAINSFOCUS";
	elseif (powerType == SPELL_POWER_ENERGY) then verboseMessage = "DRAINSENERGY";
	elseif (powerType == -2) then verboseMessage = "DRAINSHEALTH";
	elseif (powerType == SPELL_POWER_RUNES) then verboseMessage = "DRAINSRUNES";
	elseif (powerType == SPELL_POWER_RUNIC_POWER) then verboseMessage = "DRAINSRUNICPOWER";
	end
	
 -- ACTIONS TAKEN BY YOU -----------------------
 if (self:IsYou(srcName, srcFlags)) then
		if (powerType == -2) then damageGiven = self:Add(damageGiven, amount); end
		option = "you_drain";
		 
 -- ACTIONS TAKEN BY YOUR PET ------------------
 elseif (self:IsPet(srcName, srcFlags)) then
		if (powerType == -2) then petDamageGiven = self:Add(petDamageGiven, amount); end
  option = "pet_drain";
		
 -- ATTACKER VERSUS YOU OR ITSELF --------------
 elseif (self:IsYou(dstName, dstFlags)) or (self:IsTarget(srcName, srcFlags) and self:IsTarget(dstName, dstFlags)) then
		if (powerType == -2) then damageTaken = self:Add(damageTaken, amount); end
  option = "mob_drain";
		
 -- ATTACKER VERSUS YOUR PET -------------------
 elseif (self:IsPet(dstName, dstFlags)) then
		if (powerType == -2) then petDamageTaken = self:Add(petDamageTaken, amount); end
  option = "vspet_drain";

 -- OTHERS VERSUS OTHERS -----------------------
	else
  option = "oth_drain";
		
 end

 -- Output
	self:Output(option, srcName, srcFlags, dstName, dstFlags, spellName, verboseMessage, {
	 amount = amount,
	 spellSchool = spellSchool,
	 spellId = spellId,
	 powerType = powerType
	});

end


--
-- Handles extra attacks gained.
--
function HitsMode:OutputExtraAttacks(eventType, srcName, srcFlags, dstName, dstFlags, spellId, spellName, spellSchool, amount)

	self:DebugWrite(string.format(YEL.."OutputExtraAttacks received eventType: %s || src: %s || dst: %s || amount: %s || spellName: %s|r", eventType or "nil", srcName or "nil", dstName or "nil", amount or "nil", spellName or "nil"));
	if (amount == nil) then return; end
	
	local option, verboseMessage;
	
	if (self:IsYou(dstName, dstFlags)) then verboseMessage = "GAINATTACKSFROM"; else verboseMessage = "GAINSATTACKSFROM"; end
	
 -- YOU GAIN EXTRA ATTACKS ---------------------
	if (self:IsYou(dstName, dstFlags)) then
		attacksGained = self:Add(attacksGained, amount);
		option = "you_gainatk";
	
 -- YOUR PET GAINS EXTRA ATTACKS ---------------
	elseif (self:IsPet(dstName, dstFlags)) then
  option = "pet_gainatk";
	
 -- YOUR TARGET GAINS EXTRA ATTACKS ------------
	elseif (self:IsTarget(dstName, dstFlags)) then
  option = "mob_gainatk";
	
 -- SOMEONE ELSE GAINS EXTRA ATTACKS -----------
	else
  option = "oth_gainatk";
	
	end

 -- Output
	self:Output(option, srcName, srcFlags, dstName, dstFlags, spellName, verboseMessage, {
	 amount = amount,
	 spellSchool = spellSchool,
	 spellId = spellId
	});
	
end

----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

--
-- This function handles all unit filters. It returns true if the given source/destination should be displayed in the current context.
--
local overrideFilter;

function HitsMode:MatchesFilters(srcName, srcFlags, dstName, dstFlags)
	if (self:Get("enable_filters") == 0) then return true; end
	
	-- Debug section (comment out for performance)
	--self:DebugWrite(string.format("((%s AND %s) OR (%s AND %s)) AND (%s AND %s)",
		--self:FormatBool(self:SpecFilter("src", srcName, srcFlags)),
		--self:FormatBool(self:TypeFilter("src", srcName, srcFlags)),
		--self:FormatBool(self:SpecFilter("dst", dstName, dstFlags)),
		--self:FormatBool(self:TypeFilter("dst", dstName, dstFlags)),
		--self:FormatBool(self:NameFilter("src", srcName, srcFlags)),
		--self:FormatBool(self:NameFilter("dst", dstName, dstFlags))));

	-- The overall logic goes like this:
	--((SourceSpecial AND SourceType) OR (DestSpec AND DestType)) AND (SourceName AND DestName)
	return
		(
			(self:SpecFilter("src", srcName, srcFlags) and self:TypeFilter("src", srcName, srcFlags))
			or
			(self:SpecFilter("dst", dstName, dstFlags) and self:TypeFilter("dst", dstName, dstFlags))
		)
		and
		(
			self:NameFilter("src", srcName, srcFlags) and self:NameFilter("dst", dstName, dstFlags)
		);
end

function HitsMode:SpecFilter(prefix, srcName, srcFlags)
	if (self:Get("enable_flt_"..prefix.."_spec") == 0) then
		-- If the Type filter is on, we need to return true here to allow the Type filter to work
		if (self:Get("enable_flt_"..prefix.."_type") == 1) then return true; else
			-- If neither Type nor Special is on, we need to return true to cause no filtering to take place
			if (self:Get("enable_flt_"..prefix.."_spec") == 0 and self:Get("enable_flt_"..prefix.."_type") == 0) then
				-- If the opposite prefix is on, we need to return false to only allow that filter to work
				local newprefix;
				if (prefix == "src") then newprefix = "dst"; else newprefix = "src"; end
				if (self:Get("enable_flt_"..newprefix.."_spec") == 1) then return false; else
					return true;
				end
			else
				return false;
			end
		end
	end
	
	if
	(self:Get("flt_"..prefix.."_spec_me") == 1 and self:IsYou(srcName, srcFlags)) or
	(self:Get("flt_"..prefix.."_spec_mypet") == 1 and self:IsPet(srcName, srcFlags)) or
	(self:Get("flt_"..prefix.."_spec_party") == 1 and self:IsParty(srcName, srcFlags)) or
	(self:Get("flt_"..prefix.."_spec_target") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_TARGET)) or
	(self:Get("flt_"..prefix.."_spec_focus") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_FOCUS)) or
	(self:Get("flt_"..prefix.."_spec_tank") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_MAINTANK)) or
	(self:Get("flt_"..prefix.."_spec_assist") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_MAINASSIST)) or
	(self:Get("flt_"..prefix.."_spec_raidtarget") == 1 and (
		self:Is(srcFlags, COMBATLOG_OBJECT_RAIDTARGET1) or
		self:Is(srcFlags, COMBATLOG_OBJECT_RAIDTARGET2) or 
		self:Is(srcFlags, COMBATLOG_OBJECT_RAIDTARGET3) or
		self:Is(srcFlags, COMBATLOG_OBJECT_RAIDTARGET4) or
		self:Is(srcFlags, COMBATLOG_OBJECT_RAIDTARGET5) or
		self:Is(srcFlags, COMBATLOG_OBJECT_RAIDTARGET6) or
		self:Is(srcFlags, COMBATLOG_OBJECT_RAIDTARGET7) or
		self:Is(srcFlags, COMBATLOG_OBJECT_RAIDTARGET8)
	)) or
	(self:Get("flt_"..prefix.."_spec_partypets") == 1 and self:IsPartyPet(srcName, srcFlags)) or
	(self:Get("flt_"..prefix.."_spec_raid") == 1 and self:IsRaid(srcName, srcFlags)) or
	(self:Get("flt_"..prefix.."_spec_raidpets") == 1 and self:IsRaidPet(srcName, srcFlags)) then
		return true;
	else
		return false;
	end
end

function HitsMode:TypeFilter(prefix, srcName, srcFlags)
	if (self:Get("enable_flt_"..prefix.."_type") == 0) then
		-- If the Special filter is on, we need to return true here to allow the Special filter to work
		if (self:Get("enable_flt_"..prefix.."_spec") == 1) then return true; else
			-- If neither Type nor Special is on, we need to return true to cause no filtering to take place
			if (self:Get("enable_flt_"..prefix.."_spec") == 0 and self:Get("enable_flt_"..prefix.."_type") == 0) then
				-- If the opposite prefix is on, we need to return false to only allow that filter to work
				local newprefix;
				if (prefix == "src") then newprefix = "dst"; else newprefix = "src"; end
				if (self:Get("enable_flt_"..newprefix.."_type") == 1) then return false; else
					return true;
				end
			else
				return false;
			end
		end
	end
	
	if
	(self:Get("flt_"..prefix.."_affil_mine") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_AFFILIATION_MINE)) or
	(self:Get("flt_"..prefix.."_affil_party") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_AFFILIATION_PARTY)) or
	(self:Get("flt_"..prefix.."_affil_raid") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_AFFILIATION_RAID)) or
	(self:Get("flt_"..prefix.."_affil_outsider") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_AFFILIATION_OUTSIDER)) then
		if
		(self:Get("flt_"..prefix.."_reac_friendly") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_REACTION_FRIENDLY)) or
		(self:Get("flt_"..prefix.."_reac_neutral") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_REACTION_NEUTRAL)) or
		(self:Get("flt_"..prefix.."_reac_hostile") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_REACTION_HOSTILE)) then
			if
			(self:Get("flt_"..prefix.."_cont_player") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_CONTROL_PLAYER)) or
			(self:Get("flt_"..prefix.."_cont_npc") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_CONTROL_NPC)) then
				if
				(self:Get("flt_"..prefix.."_type_player") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_TYPE_PLAYER)) or
				(self:Get("flt_"..prefix.."_type_npc") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_TYPE_NPC)) or
				(self:Get("flt_"..prefix.."_type_pet") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_TYPE_PET)) or
				(self:Get("flt_"..prefix.."_type_guardian") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_TYPE_GUARDIAN)) or
				(self:Get("flt_"..prefix.."_type_object") == 1 and self:Is(srcFlags, COMBATLOG_OBJECT_TYPE_OBJECT)) then
					return true;
				end
			end
		end
	end
	return false;
end

function HitsMode:NameFilter(prefix, srcName, srcFlags)
	if (self:Get("enable_flt_"..prefix.."_name") == 0) then return true; end
	
	local i, good, filtered;
	good = false;
	filtered = false;
	
	-- Only allow these names
	for i = 1, 4 do
		if (self:Get("flt_"..prefix.."_includenames"..i) == 1) then
			filtered = true;
			if (self:IsInList(srcName, self:GetTextValue("flt_"..prefix.."_includenames"..i))) then
				good = true;
				break;
			end
		end
	end
	
	-- We didn't want to use the first filter, so set our condition to good
	if (filtered == false) then good = true; end
	
	-- Short circuit the next comparison if we already know we can't allow this name 
	if (good == false) then return false; end
	
	-- Do not allow these names
	for i = 1, 4 do
		if (self:Get("flt_"..prefix.."_excludenames"..i) == 1) then
			if (self:IsInList(srcName, self:GetTextValue("flt_"..prefix.."_excludenames"..i))) then
				good = false;
				break;
			end
		end
	end
	
	return good;
end

----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

--
-- The main output function of HitsMode: This function loops through all open windows and hands the logic
-- off to a worker method.
--
--   option:          String; The HitsMode option that controls the visibility and color of this event, e.g. "you_hit"
--   srcName:         String; The name of the source of the event
--   srcFlags:        Number (bit field); The flags of the source of the event, as given by WoW
--   dstName:         String; The name of the destination of the event
--   dstFlags:        Number (bit field); The flags of the destination of the event, as given by WoW
--   specialAttack:   String; The spell/skill name, if any.
--   verboseMessage:  String; The verbose message to display. This is for putting sentences together, e.g. in "You hit for 12", "hit for" is the verbose message.
--   p:               Table; A table of additional parameters to the output function
--
function HitsMode:Output(option, srcName, srcFlags, dstName, dstFlags, specialAttack, verboseMessage, p)
 local count;
 
 for count = 1, 4 do
  if (self:Get("enable"..count) == 1) then
   currentWindow = count;
   coerceChat = 1;
   self:DoOutput(option, srcName, srcFlags, dstName, dstFlags, specialAttack, verboseMessage, p);
   coerceChat = 0;
  end
 end
end


--
-- A lightweight output function that just prints the indicated localized string in the configured color.
-- Outputs to all windows as configured by the user.
--
function HitsMode:OutputConfiguredString(option, message)
 local count;
 
 for count = 1, 4 do
  if (self:Get("enable"..count) == 1) then
   currentWindow = count;
   coerceChat = 1;
   if (self:Get(option) == 1) then
    self:PrintW(self:GC(option)..self:GetString(message));
   end
   coerceChat = 0;
  end
 end
end

----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

function HitsMode:GetUnitColor(type, name, flags)
	if (name == self:GetString("YOURCAP") or name == self:GetString("YOUCAP")) then name = UnitName("player"); end
 local unitColor;
 if (type == "you") then unitColor = self:GC("showyou");
 elseif (type == "pet") then unitColor = self:GC("showpet");
 elseif (type == "mob") then unitColor = self:GC("showmob");
	elseif (self:Is(flags, COMBATLOG_OBJECT_CONTROL_PLAYER) and self:Is(flags, COMBATLOG_OBJECT_REACTION_HOSTILE)) then unitColor = self:GC("showoth_playerhostile");
	elseif (self:Is(flags, COMBATLOG_OBJECT_TYPE_PET) and self:Is(flags, COMBATLOG_OBJECT_CONTROL_NPC) and self:Is(flags, COMBATLOG_OBJECT_REACTION_NEUTRAL) and self:Is(flags, COMBATLOG_OBJECT_AFFILIATION_OUTSIDER)) then unitColor = self:GC("showoth_player");
	elseif (self:Is(flags, COMBATLOG_OBJECT_CONTROL_NPC)) then unitColor = self:GC("showoth_npc");
	else unitColor = self:GC("showoth_player");
	end
 if (self:Get("colorbyclass") == 1) then
		local _, class = UnitClass(name);
		if (class ~= nil) then
			if (RAID_CLASS_COLORS[class] ~= nil) then
				unitColor = self:C(RAID_CLASS_COLORS[class].r, RAID_CLASS_COLORS[class].g, RAID_CLASS_COLORS[class].b);
			end
		end
 end
 return unitColor;
end

function HitsMode:GetRaidIcon(name, flags)
 if (self:Get("showraidicons") ~= 1) then return ""; end 
 if (self:Is(flags, COMBATLOG_OBJECT_RAIDTARGET1)) then return ICON_LIST[1].."0|t"; end
 if (self:Is(flags, COMBATLOG_OBJECT_RAIDTARGET2)) then return ICON_LIST[2].."0|t"; end
 if (self:Is(flags, COMBATLOG_OBJECT_RAIDTARGET3)) then return ICON_LIST[3].."0|t"; end
 if (self:Is(flags, COMBATLOG_OBJECT_RAIDTARGET4)) then return ICON_LIST[4].."0|t"; end
 if (self:Is(flags, COMBATLOG_OBJECT_RAIDTARGET5)) then return ICON_LIST[5].."0|t"; end
 if (self:Is(flags, COMBATLOG_OBJECT_RAIDTARGET6)) then return ICON_LIST[6].."0|t"; end
 if (self:Is(flags, COMBATLOG_OBJECT_RAIDTARGET7)) then return ICON_LIST[7].."0|t"; end
 if (self:Is(flags, COMBATLOG_OBJECT_RAIDTARGET8)) then return ICON_LIST[8].."0|t"; end
 return "";
end

function HitsMode:CreateSpellLink(spellId, amount)
 if (spellId == nil) then return ""; end
 local name, rank = GetSpellInfo(spellId);
 local s = "|Hspell:"..spellId.."|h|r|cff71d5ff["..name;
 if (amount or 0) ~= 0 then
  s = s.." ("..amount..")";
 end
 s = s.."]|r|h";
 return s;
end

--
-- The worker method that actually renders the final output to the HitsMode combat logs. This
-- function handles all the colorization, flags, text layout, source/destination swaps, and other
-- logic.
--
function HitsMode:DoOutput(option, srcName, srcFlags, dstName, dstFlags, specialAttack, verboseMessage, p)

 local srcUnitType, dstUnitType, s, colorTag, optionRevised, extraValue, showSourceUnit, showDestUnit;
 local srcUnit_text, dstUnit_text, specialAttack_text, verboseMessage_text, message_text, isHeal, isDrain, isOutOfOrder;
 local srcUnit_raidIcon, dstUnit_raidIcon, amount_original, showAmount, extraSpell_text, extraSpellVerbose_text;
 
 if (p == nil) then p = {}; end
 
 optionRevised = "";
 extraValue = "";
 amount_original = p.amount;
 
 -- If we're not supposed to display this message, exit
 if (self:Get(option) ~= 1) then return; end
 
 -- Suppress the message if it's under the threshold
 local threshold = self:GetSliderValue("threshold"..currentWindow);
 if (threshold ~= 0 and p.amount ~= nil and tonumber(p.amount) ~= nil and math.abs(tonumber(p.amount)) <= threshold) then return; end
 
 -- Suppress the message if it doesn't meet the filters
 if (self:MatchesFilters(srcName, srcFlags, dstName, dstFlags) == false) then return; end
 
 -- This bit guards against some nil comparisons
 if (srcName == nil) then
  srcUnit_text = "";
  srcName = "";
 end
 if (dstName == nil) then
  dstUnit_text = "";
  dstName = "";
 end
 if (specialAttack == nil) then
  specialAttack_text = "";
  specialAttack = "";
 end
 
 -- Set a flag if this is a heal or a drain
 if (strfind(option, "heal") or strfind(option, "hot")) then isHeal = true; else isHeal = false; end
 if (strfind(option, "drain")) then isDrain = true; else isDrain = false; end
 
 -- Flag out of order events. These events show the spell after the verbose message instead of before.
 if (
		verboseMessage == "DISPEL" or verboseMessage == "DISPELS" or
		verboseMessage == "STEAL" or verboseMessage == "STEALS" or
		verboseMessage == "STARTCASTING" or verboseMessage == "STARTSCASTING" or
		verboseMessage == "FINISHCASTING" or verboseMessage == "FINISHESCASTING" or
		verboseMessage == "FAILTOCAST" or verboseMessage == "FAILSTOCAST" or
		verboseMessage == "FAILTODISPEL" or verboseMessage == "FAILSTODISPEL" or
		verboseMessage == "INTERRUPT" or verboseMessage == "INTERRUPTS" or
		verboseMessage == "GAINATTACKSFROM" or
		(p.missType ~= nil and p.missType ~= "MISS" and (specialAttack or "") ~= "") or -- Miss events (except "miss") with a special attack
		(p.powerType ~= nil and p.powerType ~= SPELL_POWER_HAPPINESS) or -- Energize events (except happiness)
		p.auraChangeType == "GAIN" -- Aura gain events
	) then
	 isOutOfOrder = true;
	else
	 isOutOfOrder = false;
	end
	
	-- Some events need their source and destination flipped after it runs through all filters
	if (
	 verboseMessage == "ISBROKENBY" or
	 (p.missType ~= nil and p.missType ~= "MISS")
	) then
	 srcName, dstName = self:Flip(srcName, dstName);
	 srcFlags, dstFlags = self:Flip(srcFlags, dstFlags);
	end
	
 -- Process the unit names and figure out what type they are
 if (self:IsYou(srcName, srcFlags)) then
  srcUnitType = "you";
		if (self:Get("renameplayer") == 1) then
			if (specialAttack ~= "" and isOutOfOrder == false and verboseMessage ~= "APPLYENCHANTMENTON") then srcName = self:GetString("YOURCAP"); else srcName = self:GetString("YOUCAP"); end
		end
 elseif (self:IsPet(srcName, srcFlags)) then
  srcUnitType = "pet";
 elseif (self:IsTarget(srcName, srcFlags)) then
  srcUnitType = "mob";
 else
		srcUnitType = "oth";
 end
 if (self:IsYou(dstName, dstFlags)) then
  dstUnitType = "you";
		if (self:Get("renameplayer") == 1) then
			dstName = self:GetString("YOUCAP");
			if (specialAttack ~= "" and srcName == "" and isOutOfOrder == false and p.missType == nil) then dstName = self:GetString("YOURCAP"); else dstName = self:GetString("YOUCAP"); end
		end
 elseif (self:IsPet(dstName, dstFlags)) then
  dstUnitType = "pet";
 elseif (self:IsTarget(dstName, dstFlags)) then
  dstUnitType = "mob";
 else
		dstUnitType = "oth";
 end
 
 -- Change message from "hits for" to "hit for" if the source is you and there is no special attack
 if (specialAttack == "" and srcUnitType == "you") then
		if (verboseMessage == "HITSFOR") then verboseMessage = "HITFOR";
		elseif (verboseMessage == "CRITSFOR") then verboseMessage = "CRITFOR";
		end
	end
	
	-- If source is null but we have a destination, flip them (this fixes events that are missing a source and weren't caught prior to this)
	if (srcName == "" and dstName ~= "") then
		srcName, dstName = self:Flip(srcName, dstName);
		srcUnitType, dstUnitType = self:Flip(srcUnitType, dstUnitType);
		srcFlags, dstFlags = self:Flip(srcFlags, dstFlags);
	end
	
	-- For environmental damage we hardcode source to "You" and destination to nothing
	if (option == "you_fall") then
		dstName = "";
		srcName = self:GetString("YOUCAP");
		srcUnitType = "you";
		dstUnitType = "you";
	end
	
 -- Get the localized verbose message text
 if (verboseMessage == nil) then verboseMessage_text = ""; else verboseMessage_text = self:GetString(verboseMessage); end
 

 -- Handlers if we have an amount to show
 if (p.amount ~= nil and p.amount ~= "") then
 
  -- Spell school coloring (don't color heals or energizes, do color drains)
  if (self:Get("schoCol") == 1) then
   if (p.spellSchool ~= nil and isHeal == false and (powerType == nil or isDrain == true)) then
    if (p.spellSchool == 0) then -- Ignored
    elseif (p.spellSchool == SCHOOL_MASK_PHYSICAL) then -- Ignored
			 elseif (p.spellSchool == SCHOOL_MASK_HOLY or p.spellSchool == (SCHOOL_MASK_PHYSICAL + SCHOOL_MASK_HOLY)) then optionRevised = "Holy";
			 elseif (p.spellSchool == SCHOOL_MASK_FIRE or p.spellSchool == (SCHOOL_MASK_PHYSICAL + SCHOOL_MASK_FIRE)) then optionRevised = "Fire";
			 elseif (p.spellSchool == SCHOOL_MASK_NATURE or p.spellSchool == (SCHOOL_MASK_PHYSICAL + SCHOOL_MASK_NATURE)) then optionRevised = "Nature";
			 elseif (p.spellSchool == SCHOOL_MASK_FROST or p.spellSchool == (SCHOOL_MASK_PHYSICAL + SCHOOL_MASK_FROST)) then optionRevised = "Frost";
			 elseif (p.spellSchool == SCHOOL_MASK_SHADOW or p.spellSchool == (SCHOOL_MASK_PHYSICAL + SCHOOL_MASK_SHADOW)) then optionRevised = "Shadow";
			 elseif (p.spellSchool == SCHOOL_MASK_ARCANE or p.spellSchool == (SCHOOL_MASK_PHYSICAL + SCHOOL_MASK_ARCANE)) then optionRevised = "Arcane";
			 elseif (p.spellSchool == SCHOOL_MASK_FIRE + SCHOOL_MASK_FROST) then optionRevised = "FrostFire";
			 elseif (p.spellSchool == SCHOOL_MASK_NATURE + SCHOOL_MASK_FROST) then optionRevised = "FrostStorm";
			 elseif (p.spellSchool == SCHOOL_MASK_NATURE + SCHOOL_MASK_SHADOW) then optionRevised = "ShadowStorm";
			 elseif (p.spellSchool == SCHOOL_MASK_FROST + SCHOOL_MASK_SHADOW) then optionRevised = "ShadowFrost";
			 elseif (p.spellSchool == SCHOOL_MASK_HOLY + SCHOOL_MASK_ARCANE) then optionRevised = "Divine";
			 elseif (p.spellSchool == SCHOOL_MASK_FIRE + SCHOOL_MASK_ARCANE) then optionRevised = "SpellFire";
			 --else self:Print("Unknown Spell School: "..p.spellSchool);
			 end
   end
  end
  
  -- Extra values
  if (self:Get("extval") == 1) then
   extraValue = "";
   if (p.resisted ~= nil) then extraValue = extraValue..string.format(" (%s %s)", p.resisted, self:GetString("RESISTED")); end
   if (p.blocked ~= nil) then extraValue = extraValue..string.format(" (%s %s)", p.blocked, self:GetString("BLOCKED")); end
   if (p.absorbed ~= nil) then extraValue = extraValue..string.format(" (%s %s)", p.absorbed, self:GetString("ABSORBED")); end
   if (p.crushing ~= nil) then extraValue = extraValue..string.format(" (%s)", self:GetString("CRUSHING")); end
   if (p.glancing ~= nil) then extraValue = extraValue..string.format(" (%s)", self:GetString("GLANCING")); end
   if ((p.overhealing or 0) > 0) then extraValue = extraValue..string.format(" (%s %s)", p.overhealing, self:GetString("OVERHEALED")); end
   if ((p.overkill or 0) > 0) then extraValue = extraValue..string.format(" (%s %s)", p.overkill, self:GetString("OVERKILL")); end
			if (extraValue ~= "") then extraValue = self:GC("extval")..extraValue.."|r"; end
  end

  -- Heal flag
  if (isHeal and self:Get("flagheals") == 1) then
   p.amount = "+"..p.amount;
  end
  
  -- Energize flag
  if (p.powerType ~= nil and self:Get("flagenergizes") == 1 and isDrain == false) then
			p.amount = "+"..p.amount;
  end
  
  -- Pet flag
  if ((self:IsPet(srcName, srcFlags) or self:IsPet(dstName, dstFlags)) and self:Get("flagpet") == 1) then
   p.amount = "~"..p.amount.."~";
  end
  
  -- If we didn't want to revise the option, go with the original
  if (optionRevised == "") then
   optionRevised = option;
  end

  -- p.critical flag and color
  if (p.critical == true) then
   if (self:Get("flagcrits") == 1) then
    p.amount = string.format("%s*|r%s%s|r%s*|r", self:GC("flagcrits"), self:GCO(optionRevised, srcUnitType, "crit"), p.amount, self:GC("flagcrits"));
   else
    p.amount = string.format("%s%s|r", self:GCO(optionRevised, srcUnitType, "crit"), p.amount);
			end
  else
   p.amount = self:GCO(optionRevised, srcUnitType, "not")..p.amount.."|r";
  end
  
  -- Append extra values if any
  message_text = p.amount..extraValue;
  showAmount = true;

 else
  showAmount = false;
 end
 
 -- Colorize spell name or convert it to link, including the rank number if this was an aura event
 if (specialAttack ~= nil and specialAttack ~= "") then
		if (p.spellId ~= nil and self:Get("showspelllinks") == 1) then -- Spell link
		 if (p.auraType ~= nil) then
			 specialAttack_text = self:CreateSpellLink(p.spellId, amount_original);
			 -- Now that we've used the amount as the spell rank, clear amount so we don't show it
		  amount_original = "";
			 p.amount = "";
			 message_text = "";
			 showAmount = false;
			else
			 specialAttack_text = self:CreateSpellLink(p.spellId);
			end
		else
		 if (p.auraType ~= nil) and (amount_original ~= nil and amount_original ~= "") then -- No link; colorize instead
		  specialAttack_text = self:GC("showskill")..specialAttack.." ("..amount_original..")|r";
			 -- Now that we've used the amount as the spell rank, clear amount so we don't show it
		  amount_original = "";
		  p.amount = "";
			 message_text = "";
			 showAmount = false;
		 else
			 specialAttack_text = self:GC("showskill")..specialAttack.."|r";
			end
		end
 else
  specialAttack_text = "";
 end
 
 -- Colorize extra spell name or convert it to link (won't have a rank number)
 if (p.extraSpell ~= nil) then
  if (p.extraSpellId ~= nil and self:Get("showspelllinks") == 1) then
   extraSpell_text = self:CreateSpellLink(p.extraSpellId);
  else
   extraSpell_text = self:GC("showskill")..p.extraSpell.."|r";
  end
 else
  extraSpell_text = nil;
 end
 
 -- Colorize unit names
 srcUnit_text = string.format("%s%s%s|r%s", self:BracketLeft(), self:GetUnitColor(srcUnitType, srcName, srcFlags), srcName, self:BracketRight());
 dstUnit_text = string.format("%s%s%s|r%s", self:BracketLeft(), self:GetUnitColor(dstUnitType, dstName, dstFlags), dstName, self:BracketRight());
 
 -- Get raid icons
 srcUnit_raidIcon = self:GetRaidIcon(srcName, srcFlags);
 dstUnit_raidIcon = self:GetRaidIcon(dstName, dstFlags);
 
 -- Determine if we should show the source and destination unit names
 if (srcUnitType == "you" and self:Get("showyou") == 1) or
    (srcUnitType == "pet" and self:Get("showpet") == 1) or
    (srcUnitType == "mob" and self:Get("showmob") == 1) or
    (srcUnitType == "oth" and (
					(self:Is(srcFlags, COMBATLOG_OBJECT_CONTROL_PLAYER) and self:Is(srcFlags, COMBATLOG_OBJECT_REACTION_HOSTILE) and self:Get("showoth_playerhostile") == 1) or
					(self:Is(srcFlags, COMBATLOG_OBJECT_CONTROL_PLAYER) and self:Get("showoth_player") == 1) or
					(self:Get("showoth_npc") == 1)
				)) then showSourceUnit = true; else showSourceUnit = false; end
 if (dstUnitType == "you" and self:Get("showyou") == 1) or
    (dstUnitType == "pet" and self:Get("showpet") == 1) or
    (dstUnitType == "mob" and self:Get("showmob") == 1) or
    (dstUnitType == "oth" and (
					(self:Is(dstFlags, COMBATLOG_OBJECT_CONTROL_PLAYER) and self:Is(dstFlags, COMBATLOG_OBJECT_REACTION_HOSTILE) and self:Get("showoth_playerhostile") == 1) or
					(self:Is(dstFlags, COMBATLOG_OBJECT_CONTROL_PLAYER) and self:Get("showoth_player") == 1) or
					(self:Get("showoth_npc") == 1)
				)) then showDestUnit = true; else showDestUnit = false; end
	 
 -- If we're showing an extra spell, get the verbose message for it
 if (p.extraSpell_verboseMessage ~= nil) then
  extraSpellVerbose_text = self:GetString(p.extraSpell_verboseMessage);
 else
  extraSpellVerbose_text = "";
 end

 -- Colorize verboseMessage and extraSpellVerbose_text
 if (self:Get("verboseb") == 1) then
  colorTag = self:GC("verboseb");
 else
  colorTag = self:C(1, 1, 1);
 end
 verboseMessage_text = colorTag..verboseMessage_text.."|r";
 if (p.extraSpell_verboseMessage ~= nil) then extraSpellVerbose_text = colorTag..extraSpellVerbose_text.."|r"; end

 -- Fix a missing unit name for some out of order messages
 --if (isOutOfOrder) then
		--if (dstName == "" and srcName ~= "") then
			--dstUnit_text = srcUnit_text;
		--elseif (srcName == "" and dstName ~= "") then
			--srcUnit_text = dstUnit_text;
		--end
 --end

	-- No unit name available; hide it
	if (srcName == nil or srcName == "") then showSourceUnit = false; end
	if (dstName == nil or dstName == "") then showDestUnit = false; end
	
	-- Final check for source/dest sanity. If we have a dest but no source at this point, and we are in an event which might cause this, flip source/dest.
	if (isOutOfOrder and showSourceUnit == false and showDestUnit == true) then
		srcName, dstName = self:Flip(srcName, dstName);
		srcUnitType, dstUnitType = self:Flip(srcUnitType, dstUnitType);
		srcFlags, dstFlags = self:Flip(srcFlags, dstFlags);
		showSourceUnit, showDestUnit = self:Flip(showSourceUnit, showDestUnit);
	end
	
	
	-- Debug output
	if (g_debugWrote == true) then self:DebugWrite2(HIL.."> verboseMessage: "..(verboseMessage or "<none>").." | p.amount: "..(amount_original or "<none>").." | specialAttack: "..(specialAttack or "<none>").." | p.missType: "..(p.missType or "<none>").." | isOutOfOrder: "..self:FormatBool(isOutOfOrder).." | showAmount: "..self:FormatBool(showAmount).." | Src: "..srcName.." | Dst: "..dstName.." | isHeal: "..self:FormatBool(isHeal).." | ShowSrc: "..self:FormatBool(showSourceUnit).." | ShowDst: "..self:FormatBool(showDestUnit).." | p.auraType: "..(p.auraType or "<none>").."|r"); end


 ------
 --- Main output
 ------
 
 -------- VERBOSE MODE
 if (self:Get("verboseb") == 1) then

  s = "";
  
  -- Show value before message if turned on
  if (self:Get("verbosec") == 1) then
		 if (showAmount) then
			 s = s..message_text..self:GC("verboseb")..self:GetString("CDIVIDER").."|r ";
		 end
  end

  s = s..srcUnit_raidIcon;
  if (showSourceUnit) then s = s..srcUnit_text.." "; end
  if (isOutOfOrder) then s = s..verboseMessage_text.." "; end
  if (specialAttack ~= nil and specialAttack ~= "" and self:Get("showskill") == 1) then s = s..specialAttack_text.." "; end
  if (isOutOfOrder == false) then s = s..verboseMessage_text.." "; end
  if (p.showOn == true) then s = s..self:GetString("ON").." "; end
  s = s..dstUnit_raidIcon;
		if (showDestUnit) then s = s..dstUnit_text.." "; end
		
		-- Show value after message if configured
  if (self:Get("verbosec") ~= 1) then
		 if (showAmount) then
		  -- Show "for" if turned on
			 if (self:Get("shortmode") ~= 1 and (((srcName ~= "" and dstName ~= "") or p.powerType ~= nil) and tonumber(amount_original) ~= nil)) then s = s..self:GC("verboseb")..self:GetString("FOR").."|r "; end
			 s = s..message_text.." ";
		 end
		end

  -- Extra spell
  if (p.extraSpell_verboseMessage ~= nil) then s = s..extraSpellVerbose_text; end
  if (p.extraSpell ~= nil) then s = s.." "..extraSpell_text; end
  if (p.extraSpellId ~= nil and self:Get("showicons") == 1) then
		 local spellTexture = self:GetSpellIcon(p.extraSpellId);
		 if spellTexture ~= nil then s = s.." |T"..spellTexture..":0|t"; end
  end

 -------- NO VERBOSE MODE
 else
  s = "";
		if (self:Get("showskill") == 1) then s = specialAttack_text.." "; end
		s = s..dstUnit_raidIcon;
		if (showAmount) then s = s..message_text; end
		if (p.missType ~= nil) then s = s..verboseMessage_text; end
 end
 
 -- Add the spell texture if turned on
 if (p.spellId ~= nil and self:Get("showicons") == 1) then
		local spellTexture = self:GetSpellIcon(p.spellId);
		if spellTexture ~= nil then s = "|T"..spellTexture..":0|t "..s; end
 end
 
 -- Output the text
 if (s ~= "" and s ~= " ") then self:PrintW(s); end
end



-- Iterate through all the open windows and output the combat summary
function HitsMode:SummarizeCombat()
 if (self:Get("enable_combatsummary") == 0) then return; end
 local endTime = GetTime();
 local seconds = endTime - combatStartTime;
 if (seconds == 0) then return; end
 local count;
 for count = 1, 4 do
  if (self:Get("enable"..count) == 1) then
   currentWindow = count;
   coerceChat = 1;
   self:DoSummarizeCombat(seconds);
   coerceChat = 0;
  end
 end
end


function HitsMode:Announce(text, r, g, b)
	if tostring(SHOW_COMBAT_TEXT) ~= "0" then
		CombatText_AddMessage(text, CombatText_StandardScroll, r, g, b, nil, false)
	else
		UIErrorsFrame:AddMessage(text, r, g, b, 1, UIERRORS_HOLD_TIME)
	end
end


-- Handles the actual output of the combat summary
function HitsMode:DoSummarizeCombat(seconds)
 if (seconds <= 0) then return; end
 if (self:Get("cs_skipentire") == 1 and damageGiven == 0 and petDamageGiven == 0 and healsGiven == 0 and expGained == 0) then return; end
 if (self:Get("cs_sep_short") == 1) then self:PrintW(self:GC("cs_sep_short")..self:GetString("SHORTSEP").."|r"); end
 if (self:Get("cs_sep1") == 1) then self:PrintW(self:GC("cs_sep1").."~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|r"); end
 if (self:Get("cs_duration") == 1) then self:PrintW(self:GC("cs_duration").."Fight duration: |r"..self:GC("cs_numeric")..SecondsToTime(seconds).."|r"); end
 self:OutputCombatSummary(L[294], damageGiven + petDamageGiven, seconds, "skip_yourdamage", "cs_totaldamagegiven"); -- Total damage given:
 self:OutputCombatSummary(L[186], damageGiven, seconds, "skip_yourdamage", "cs_damagegiven"); -- Damage given (you):
 self:OutputCombatSummary(L[187], damageResisted, seconds, "skip_avoided", "cs_damageavoided"); -- Damage avoided (you): 
 self:OutputCombatSummary(L[188], damageTaken, seconds, "skip_yourdamage", "cs_damagetaken"); -- Damage taken (you): 
 self:OutputCombatSummary(L[189], petDamageGiven, seconds, "skip_petdamage", "cs_petdamagegiven"); -- Damage given (your pet): 
 self:OutputCombatSummary(L[190], petDamageTaken, seconds, "skip_petdamage", "cs_petdamagetaken"); -- Damage taken (your pet): 
 self:OutputCombatSummary(L[191], healsGiven, seconds, "skip_yourheals", "cs_healsgiven"); -- Heals given: 
 self:OutputCombatSummary(L[192], healsTaken, seconds, "skip_yourheals", "cs_healsreceived"); -- Heals received (you): 
 self:OutputCombatSummary(L[193], petHealsTaken, seconds, "skip_yourheals", "cs_pethealsreceived"); -- Heals received (your pet): 
 self:OutputCombatSummary(L[194], manaGained, seconds, "1", "cs_gained"); -- Mana gained: 
 self:OutputCombatSummary(L[195], energyGained, seconds, "1", "cs_gained"); -- Energy gained: 
 self:OutputCombatSummary(L[196], rageGained, seconds, "1", "cs_gained"); -- Rage gained: 
 self:OutputCombatSummary(L[197], attacksGained, seconds, "1", "cs_attacksgained"); -- Attacks gained: 
 self:OutputCombatSummary(L[199], skillsIncreased, seconds, "1", "cs_skills"); -- Skills increased: 
 self:OutputCombatSummary(L[200], expGained, seconds, "1", "cs_exp"); -- Exp gained: 
 if (self:Get("cs_sep2") == 1) then self:PrintW(self:GC("cs_sep2").."~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|r"); end
 local dps = (damageGiven + petDamageGiven) / seconds;
 if (dps <= 0 or dps >= 5000) then return; end
 local r, g, b = self:GetColor("cs_announcedps");
 if (announced == false) then
  if (self:Get("cs_announcedps") == 1) then self:Announce(L[301]..self:FormatNumber(dps), r, g, b); end
  if (pOptions.maxdps == nil) then pOptions.maxdps = 0.0; end
  if (dps > pOptions.maxdps) then
   local r, g, b = self:GetColor("cs_announcehighdps");
   pOptions.maxdps = dps;
   if (self:Get("cs_announcehighdps") == 1) then self:Announce(L[299]..self:FormatNumber(dps)..L[300], r, g, b); end
  end
  announced = true;
 end
end


function HitsMode:OutputCombatSummary(text, value, seconds, skipOption, messageOption)
 -- Skip the message if it's been turned off
 if (self:Get(messageOption) ~= 1) then return; end
 
 value = math.abs(value);

 -- Get whether to auto-skip or not
 local skip = 0;
 if (skipOption == "1") then
  skip = 1;
 else
  skip = self:Get(skipOption, 1);
 end
 
 -- Get the configured colors of the message
 local color = self:GC(messageOption);
 local numcolor = self:GC("cs_numeric");
 
 -- Output the message if necessary
 if ((skip == 1 and value > 0) or skip == 0) then
  local s = color..text.."|r"..numcolor..self:FormatNumber(value).."|r "..color.."(|r"..numcolor..self:FormatNumber(value / seconds).."|r";
  if (self:Get("cs_sec") == 1) then s = s..self:GC("cs_sec").."/sec".."|r"; end
  s = s..color..")|r";
  self:PrintW(s);
 end
end


function HitsMode:AutoSkip(amount, option)
 local skip = self:Get(option, 1);
 if ((amount > 0 and skip == 1) or skip == 0) then
  return 1;
 else
  return 0;
 end;
end
